from allrank.click_models.base import FixedClickModel
from tests.click_models import click


def test_fixed_click_model_single():
    click_model = FixedClickModel([0])
    assert click(click_model, [], [1]) == [1]
    assert click(click_model, [], [1, 2]) == [1, 0]
    assert click(click_model, [], [1, 2, 3]) == [1, 0, 0]


def test_fixed_click_model_multiple():
    assert click(FixedClickModel([0, 1]), [], [1, 2, 3, 4]) == [1, 1, 0, 0]
    assert click(FixedClickModel([0, 1, 2]), [], [1, 2, 3, 4]) == [1, 1, 1, 0]
    assert click(FixedClickModel([0, 2, 3]), [], [1, 2, 3, 4]) == [1, 0, 1, 1]



import os
import tempfile
from typing import Any
from urllib.parse import urlparse

import gcsfs
from attr import attrib, attrs
from pkg_resources import Requirement, resource_filename

from allrank.utils.command_executor import execute_command
from allrank.utils.ltr_logging import get_logger

logger = get_logger()


@attrs
class PathsContainer:
    local_base_output_path = attrib(type=str)
    base_output_path = attrib(type=str)
    output_dir = attrib(type=str)
    tensorboard_output_path = attrib(type=str)
    config_path = attrib(type=str)

    @classmethod
    def from_args(cls, output, run_id, config_path, package_name="allrank"):
        base_output_path = get_path_from_local_uri(output)
        if is_gs_path(base_output_path):
            local_base_output_path = tempfile.mkdtemp()
        else:
            local_base_output_path = base_output_path
        output_dir = os.path.join(local_base_output_path, "results", run_id)
        tensorboard_output_path = os.path.join(base_output_path, "tb_evals", "single", run_id)
        if not os.path.exists(config_path):
            print("config not exists at {}, extracting config file path from package {}".format(config_path, package_name))
            config_path = resource_filename(Requirement.parse(
                package_name), os.path.join(package_name, config_path))
        print("will read config from {}".format(config_path))
        return cls(local_base_output_path, base_output_path, output_dir, tensorboard_output_path, config_path)


def clean_up(path):
    rm_command = "rm -rf {path}".format(path=path)
    execute_command(rm_command)


def create_output_dirs(output_path: str) -> None:
    for subdir in ["models", "models/partial", "evals", "evals/tensorboard", "predictions"]:
        os.makedirs(os.path.join(output_path, subdir), exist_ok=True)


def get_path_from_local_uri(uri: Any) -> str:
    parsed = urlparse(uri)
    if parsed.scheme == "file":
        return parsed.netloc + parsed.path
    else:
        return uri


def is_gs_path(uri) -> bool:
    return urlparse(uri).scheme == "gs"


def open_local_or_gs(path, mode):
    open_func = gcsfs.GCSFileSystem().open if is_gs_path(path) else open
    return open_func(path, mode)


def copy_file_to_local(uri: str) -> str:
    temp_dir = tempfile.mkdtemp()
    local_file = "local_file"
    command = "gsutil cp {gs_uri} {local_path}".format(
        gs_uri=uri, local_path=os.path.join(temp_dir, local_file))
    execute_command(command)
    return os.path.join(temp_dir, local_file)


def copy_local_to_gs(source_local: str, destination_uri: str) -> None:
    command = "gsutil cp -r {source_local}/* {destination_uri}".format(
        source_local=source_local, destination_uri=destination_uri)
    execute_command(command)



"""
Code in this file was adapted from "The Annotated Transformer" by Harvard NLP.
http://nlp.seas.harvard.edu/2018/04/03/attention.html
"""

import copy
import math
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from allrank.config import PositionalEncoding
from allrank.models.positional import _make_positional_encoding


def clones(module, N):
    """
    Creation of N identical layers.
    :param module: module to clone
    :param N: number of copies
    :return: nn.ModuleList of module copies
    """
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class Encoder(nn.Module):
    """
    Stack of Transformer encoder blocks with positional encoding.
    """
    def __init__(self, layer, N, position):
        """
        :param layer: single building block to clone
        :param N: number of copies
        :param position: positional encoding module
        """
        super(Encoder, self).__init__()
        self.layers = clones(layer, N)
        self.norm = LayerNorm(layer.size)
        self.position = position

    def forward(self, x, mask, indices):
        """
        Forward pass through each block of the Transformer.
        :param x: input of shape [batch_size, slate_length, input_dim]
        :param mask: padding mask of shape [batch_size, slate_length]
        :param indices: original item ranks used in positional encoding, shape [batch_size, slate_length]
        :return: output of shape [batch_size, slate_length, output_dim]
        """
        if self.position:
            x = self.position(x, mask, indices)
        mask = mask.unsqueeze(-2)
        for layer in self.layers:
            x = layer(x, mask)
        return self.norm(x)


class LayerNorm(nn.Module):
    """
    Layer normalization module.
    """
    def __init__(self, features, eps=1e-6):
        """
        :param features: shape of normalized features
        :param eps: epsilon used for standard deviation
        """
        super(LayerNorm, self).__init__()
        self.a_2 = nn.Parameter(torch.ones(features))  # type: ignore
        self.b_2 = nn.Parameter(torch.zeros(features))  # type: ignore
        self.eps = eps

    def forward(self, x):
        """
        Forward pass through the layer normalization.
        :param x: input of shape [batch_size, slate_length, input_dim]
        :return: normalized input of shape [batch_size, slate_length, output_dim]
        """
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.a_2 * (x - mean) / (std + self.eps) + self.b_2


class SublayerConnection(nn.Module):
    """
    Residual connection followed by layer normalization.
    Please not that for code simplicity the norm is first as opposed to last.
    """
    def __init__(self, size, dropout):
        """
        :param size: number of input/output features
        :param dropout: dropout probability
        """
        super(SublayerConnection, self).__init__()
        self.norm = LayerNorm(size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):
        """
        Forward pass through the sublayer connection module, applying the residual connection to any sublayer with the same size.
        :param x: input of shape [batch_size, slate_length, input_dim]
        :param sublayer: layer through which to pass the input prior to applying the sum
        :return: output of shape [batch_size, slate_length, output_dim]
        """
        return x + self.dropout(
            sublayer(self.norm(x)))


class EncoderLayer(nn.Module):
    """
    Single Transformer encoder block made of self-attention and feed-forward layers with residual connections.
    """
    def __init__(self, size, self_attn, feed_forward, dropout):
        """
        :param size: input/output size of the encoder block
        :param self_attn: self-attention layer
        :param feed_forward: feed-forward layer
        :param dropout: dropout probability
        """
        super(EncoderLayer, self).__init__()
        self.self_attn = self_attn
        self.feed_forward = feed_forward
        self.sublayer = clones(SublayerConnection(size, dropout), 2)
        self.size = size

    def forward(self, x, mask):
        """
        Forward pass through the encoder block.
        :param x: input of shape [batch_size, slate_length, self.size]
        :param mask: padding mask of shape [batch_size, slate_length]
        :return: output of shape [batch_size, slate_length, self.size]
        """
        x = self.sublayer[0](x, lambda x: self.self_attn(x, x, x, mask))
        return self.sublayer[1](x, self.feed_forward)


def attention(query, key, value, mask=None, dropout=None):
    """
    Basic function for "Scaled Dot Product Attention" computation.
    :param query: query set of shape [batch_size, slate_size, n_attention_heads, attention_dim]
    :param key: key set of shape [batch_size, slate_size, n_attention_heads, attention_dim]
    :param value: value set of shape [batch_size, slate_size, n_attention_heads, attention_dim]
    :param mask: padding mask of shape [batch_size, slate_length]
    :param dropout: dropout probability
    :return: attention scores of shape [batch_size, slate_size, n_attention_heads, attention_dim]
    """
    d_k = query.size(-1)
    scores = torch.matmul(query, key.transpose(-2, -1)) / math.sqrt(d_k)

    if mask is not None:
        scores = scores.masked_fill(mask == 1, float("-inf"))

    p_attn = F.softmax(scores, dim=-1)
    if dropout is not None:
        p_attn = dropout(p_attn)
    return torch.matmul(p_attn, value), p_attn


class MultiHeadedAttention(nn.Module):
    """
    Multi-headed attention block.
    """
    def __init__(self, h, d_model, dropout=0.1):
        """
        :param h: number of attention heads
        :param d_model: input/output dimensionality
        :param dropout: dropout probability
        """
        super(MultiHeadedAttention, self).__init__()
        assert d_model % h == 0
        # We assume d_v always equals d_k
        self.d_k = d_model // h
        self.h = h
        self.linears = clones(nn.Linear(d_model, d_model), 4)
        self.attn = None
        self.dropout = nn.Dropout(p=dropout)

    def forward(self, query, key, value, mask=None):
        """
        Forward pass through the multi-head attention block.
        :param query: query set of shape [batch_size, slate_size, self.d_model]
        :param key: key set of shape [batch_size, slate_size, self.d_model]
        :param value: value set of shape [batch_size, slate_size, self.d_model]
        :param mask: padding mask of shape [batch_size, slate_length]
        :return: output of shape [batch_size, slate_size, self.d_model]
        """
        if mask is not None:
            # same mask applied to all h heads
            mask = mask.unsqueeze(1)
        nbatches = query.size(0)

        # 1) Do all the linear projections in batch from d_model => h x d_k
        query, key, value = \
            [linear(x).view(nbatches, -1, self.h, self.d_k).transpose(1, 2)
             for linear, x in zip(self.linears, (query, key, value))]

        # 2) Apply attention on all the projected vectors in batch.
        x, self.attn = attention(query, key, value, mask=mask, dropout=self.dropout)

        # 3) "Concat" using a view and apply a final linear.
        x = x.transpose(1, 2).contiguous() \
            .view(nbatches, -1, self.h * self.d_k)
        return self.linears[-1](x)


class PositionwiseFeedForward(nn.Module):
    """
    Feed-forward block.
    """
    def __init__(self, d_model, d_ff, dropout=0.1):
        """
        :param d_model: input/output dimensionality
        :param d_ff: hidden dimensionality
        :param dropout: dropout probability
        """
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        """
        Forward pass through the feed-forward block.
        :param x: input of shape [batch_size, slate_size, self.d_model]
        :return: output of shape [batch_size, slate_size, self.d_model]
        """
        return self.w_2(self.dropout(F.relu(self.w_1(x))))


def make_transformer(N=6, d_ff=2048, h=8, dropout=0.1, n_features=136,
                     positional_encoding: Optional[PositionalEncoding] = None):
    """
    Helper function for instantiating Transformer-based Encoder.
    :param N: number of Transformer blocks
    :param d_ff: hidden dimensionality of the feed-forward layer in the Transformer block
    :param h: number of attention heads
    :param dropout: dropout probability
    :param n_features: number of input/output features of the feed-forward layer
    :param positional_encoding: config.PositionalEncoding object containing PE config
    :return: Transformer-based Encoder with given hyperparameters
    """
    c = copy.deepcopy
    attn = MultiHeadedAttention(h, n_features, dropout)

    ff = PositionwiseFeedForward(n_features, d_ff, dropout)
    position = _make_positional_encoding(n_features, positional_encoding)
    return Encoder(EncoderLayer(n_features, c(attn), c(ff), dropout), N, position)



import numpy as np

from allrank.click_models.base import OnlyRelevantClickModel
from tests.click_models import click


def test_only_relevant_click_model():
    click_model = OnlyRelevantClickModel(1)
    assert click(click_model, np.array([[0, 1]]), [1]) == [1]
    assert click(click_model, np.array([[0, 1]]), [0]) == [0]
    assert click(click_model, np.array([[1, 1], [1, 0], [0, 0]]), [1, 0, 0]) == [1, 0, 0]


def test_only_relevant_above_threshold_click_model():
    click_model = OnlyRelevantClickModel(2)
    assert click(click_model, np.array([[0, 1]]), [2]) == [1]
    assert click(click_model, np.array([[0, 1]]), [1]) == [0]
    assert click(click_model, np.array([[0, 1]]), [0]) == [0]
    assert click(click_model, np.array([[1, 1], [1, 0], [0, 0]]), [0, 1, 2]) == [0, 0, 1]



import torch
import numpy as np

from allrank.models.losses import DEFAULT_EPS
from allrank.models.model_utils import get_torch_device


def sinkhorn_scaling(mat, mask=None, tol=1e-6, max_iter=50):
    """
    Sinkhorn scaling procedure.
    :param mat: a tensor of square matrices of shape N x M x M, where N is batch size
    :param mask: a tensor of masks of shape N x M
    :param tol: Sinkhorn scaling tolerance
    :param max_iter: maximum number of iterations of the Sinkhorn scaling
    :return: a tensor of (approximately) doubly stochastic matrices
    """
    if mask is not None:
        mat = mat.masked_fill(mask[:, None, :] | mask[:, :, None], 0.0)
        mat = mat.masked_fill(mask[:, None, :] & mask[:, :, None], 1.0)

    for _ in range(max_iter):
        mat = mat / mat.sum(dim=1, keepdim=True).clamp(min=DEFAULT_EPS)
        mat = mat / mat.sum(dim=2, keepdim=True).clamp(min=DEFAULT_EPS)

        if torch.max(torch.abs(mat.sum(dim=2) - 1.)) < tol and torch.max(torch.abs(mat.sum(dim=1) - 1.)) < tol:
            break

    if mask is not None:
        mat = mat.masked_fill(mask[:, None, :] | mask[:, :, None], 0.0)

    return mat


def deterministic_neural_sort(s, tau, mask):
    """
    Deterministic neural sort.
    Code taken from "Stochastic Optimization of Sorting Networks via Continuous Relaxations", ICLR 2019.
    Minor modifications applied to the original code (masking).
    :param s: values to sort, shape [batch_size, slate_length]
    :param tau: temperature for the final softmax function
    :param mask: mask indicating padded elements
    :return: approximate permutation matrices of shape [batch_size, slate_length, slate_length]
    """
    dev = get_torch_device()

    n = s.size()[1]
    one = torch.ones((n, 1), dtype=torch.float32, device=dev)
    s = s.masked_fill(mask[:, :, None], -1e8)
    A_s = torch.abs(s - s.permute(0, 2, 1))
    A_s = A_s.masked_fill(mask[:, :, None] | mask[:, None, :], 0.0)

    B = torch.matmul(A_s, torch.matmul(one, torch.transpose(one, 0, 1)))

    temp = [n - m + 1 - 2 * (torch.arange(n - m, device=dev) + 1) for m in mask.squeeze(-1).sum(dim=1)]
    temp = [t.type(torch.float32) for t in temp]
    temp = [torch.cat((t, torch.zeros(n - len(t), device=dev))) for t in temp]
    scaling = torch.stack(temp).type(torch.float32).to(dev)  # type: ignore

    s = s.masked_fill(mask[:, :, None], 0.0)
    C = torch.matmul(s, scaling.unsqueeze(-2))

    P_max = (C - B).permute(0, 2, 1)
    P_max = P_max.masked_fill(mask[:, :, None] | mask[:, None, :], -np.inf)
    P_max = P_max.masked_fill(mask[:, :, None] & mask[:, None, :], 1.0)
    sm = torch.nn.Softmax(-1)
    P_hat = sm(P_max / tau)
    return P_hat


def sample_gumbel(samples_shape, device, eps=1e-10) -> torch.Tensor:
    """
    Sampling from Gumbel distribution.
    Code taken from "Stochastic Optimization of Sorting Networks via Continuous Relaxations", ICLR 2019.
    Minor modifications applied to the original code (masking).
    :param samples_shape: shape of the output samples tensor
    :param device: device of the output samples tensor
    :param eps: epsilon for the logarithm function
    :return: Gumbel samples tensor of shape samples_shape
    """
    U = torch.rand(samples_shape, device=device)
    return -torch.log(-torch.log(U + eps) + eps)


def stochastic_neural_sort(s, n_samples, tau, mask, beta=1.0, log_scores=True, eps=1e-10):
    """
    Stochastic neural sort. Please note that memory complexity grows by factor n_samples.
    Code taken from "Stochastic Optimization of Sorting Networks via Continuous Relaxations", ICLR 2019.
    Minor modifications applied to the original code (masking).
    :param s: values to sort, shape [batch_size, slate_length]
    :param n_samples: number of samples (approximations) for each permutation matrix
    :param tau: temperature for the final softmax function
    :param mask: mask indicating padded elements
    :param beta: scale parameter for the Gumbel distribution
    :param log_scores: whether to apply the logarithm function to scores prior to Gumbel perturbation
    :param eps: epsilon for the logarithm function
    :return: approximate permutation matrices of shape [n_samples, batch_size, slate_length, slate_length]
    """
    dev = get_torch_device()

    batch_size = s.size()[0]
    n = s.size()[1]
    s_positive = s + torch.abs(s.min())
    samples = beta * sample_gumbel([n_samples, batch_size, n, 1], device=dev)
    if log_scores:
        s_positive = torch.log(s_positive + eps)

    s_perturb = (s_positive + samples).view(n_samples * batch_size, n, 1)
    mask_repeated = mask.repeat_interleave(n_samples, dim=0)

    P_hat = deterministic_neural_sort(s_perturb, tau, mask_repeated)
    P_hat = P_hat.view(n_samples, batch_size, n, n)
    return P_hat



import math

from functools import partial
from pytest import approx

from allrank.data.dataset_loading import PADDED_Y_VALUE
from tests.losses.utils import neuralNDCG_wrap, ndcg_wrap


test_cases = [{"stochastic": False, "transposed": False},
              {"stochastic": True, "transposed": False},
              {"stochastic": False, "transposed": True},
              {"stochastic": True, "transposed": True}]


def test_neuralNDCG_simple():
    for tc in test_cases:
        neuralNDCG_simple(partial(neuralNDCG_wrap, **tc))


def neuralNDCG_simple(fun):
    y_pred = [0.5, 0.2]
    y_true = [1.0, 0.0]

    result = fun(y_pred, y_true)
    expected = ndcg_wrap(y_pred, y_true)

    assert math.isfinite(result)
    assert (-1 * result == approx(expected))


def test_neuralNDCG_longer():
    for tc in test_cases:
        neuralNDCG_longer(partial(neuralNDCG_wrap, **tc))


def neuralNDCG_longer(fun):
    y_pred = [0.5, 0.2, 0.1, 0.4, 1.0, -1.0, 0.63]
    y_true = [1.0, 2.0, 2.0, 4.0, 1.0, 4.0, 3.0]

    result = fun(y_pred, y_true)
    expected = ndcg_wrap(y_pred, y_true)

    assert math.isfinite(result)
    assert (-1 * result == approx(expected))


def test_neuralNDCG_stable_for_very_small_prediction():
    for tc in test_cases:
        neuralNDCG_stable_for_very_small_prediction(partial(neuralNDCG_wrap, **tc))


def neuralNDCG_stable_for_very_small_prediction(fun):
    y_pred = [0.5, -1e30]
    y_true = [1.0, 0.0]

    result = fun(y_pred, y_true)
    expected = ndcg_wrap(y_pred, y_true)

    assert math.isfinite(result)
    assert (-1 * result == approx(expected))


def test_neuralNDCG_ignores_padded_value():
    for tc in test_cases:
        neuralNDCG_ignores_padded_value(partial(neuralNDCG_wrap, **tc))


def neuralNDCG_ignores_padded_value(fun):
    y_pred = [0.5, 0.2, 0.1, 0.4, 1.0, -1.0, 0.63, 1., 0.5, 0.3]
    y_true = [1.0, 2.0, 2.0, 4.0, 1.0, 4.0, 3.0, PADDED_Y_VALUE, PADDED_Y_VALUE, PADDED_Y_VALUE]

    result = fun(y_pred, y_true, temperature=0.001)
    expected = ndcg_wrap(y_pred, y_true)

    assert math.isfinite(result)
    assert (-1 * result == approx(expected))


def test_neuralNDCG_at_3():
    for tc in test_cases:
        neuralNDCG_at_3(partial(neuralNDCG_wrap, **tc))


def neuralNDCG_at_3(fun):
    y_pred = [0.5, 0.2, 0.1, 0.4, 1.0, -1.0, 0.63]
    y_true = [1.0, 2.0, 2.0, 4.0, 1.0, 4.0, 3.0]
    ats = 3

    result = fun(y_pred, y_true, k=ats)
    expected = ndcg_wrap(y_pred, y_true, ats=[ats])

    assert math.isfinite(result)
    assert (-1 * result == approx(expected))



from argparse import ArgumentParser
from typing import Any, Dict
from typing import Tuple

import os
import numpy as np
from sklearn.datasets import dump_svmlight_file


def generate_dummy_data(
        num_queries: int, results_len: int, num_labels: int, num_features: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Generate dummy dataset to be dumped in libsvm format.
    """
    X = np.random.randn(num_queries * results_len, num_features)
    y = np.maximum(0, (((X + 1) / 2).mean(axis=-1) * num_labels).astype(np.int32))
    qid = np.repeat(np.arange(0, num_queries), results_len)
    return X, y, qid


def parse_args() -> Dict[str, Any]:
    parser = ArgumentParser("Dummy data")
    parser.add_argument("--num_queries", help="Number of queries.", default=100)
    parser.add_argument("--results_len", help="Length of results list for a single query.", default=20)
    parser.add_argument("--num_labels", help="Number of relevance levels.", default=5)
    parser.add_argument("--num_features", help="Number of features of a single item", default=20)
    return vars(parser.parse_args())


if __name__ == '__main__':
    np.random.seed(42)
    args = parse_args()
    X_train, y_train, qid_train = generate_dummy_data(**args)
    X_val, y_val, qid_val = generate_dummy_data(**args)

    os.makedirs("dummy_data", exist_ok=True)
    dump_svmlight_file(X_train, y_train, os.path.join("dummy_data", "train.txt"), query_id=qid_train)
    dump_svmlight_file(X_val, y_val, os.path.join("dummy_data", "vali.txt"), query_id=qid_val)



import torch

from allrank.data.dataset_loading import PADDED_Y_VALUE


def pointwise_rmse(y_pred, y_true, no_of_levels, padded_value_indicator=PADDED_Y_VALUE):
    """
    Pointwise RMSE loss.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param no_of_levels: number of unique ground truth values
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: loss value, a torch.Tensor
    """
    y_pred = y_pred.clone()
    y_true = y_true.clone()

    mask = y_true == padded_value_indicator
    valid_mask = (y_true != padded_value_indicator).type(torch.float32)

    y_true[mask] = 0
    y_pred[mask] = 0

    errors = (y_true - no_of_levels * y_pred)

    squared_errors = errors ** 2

    mean_squared_errors = torch.sum(squared_errors, dim=1) / torch.sum(valid_mask, dim=1)

    rmses = torch.sqrt(mean_squared_errors)

    return torch.mean(rmses)



import os
from functools import partial

import numpy as np
import torch
from torch.nn.utils import clip_grad_norm_

import allrank.models.metrics as metrics_module
from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.model_utils import get_num_params, log_num_params
from allrank.training.early_stop import EarlyStop
from allrank.utils.ltr_logging import get_logger
from allrank.utils.tensorboard_utils import TensorboardSummaryWriter

logger = get_logger()


def loss_batch(model, loss_func, xb, yb, indices, gradient_clipping_norm, opt=None):
    mask = (yb == PADDED_Y_VALUE)
    loss = loss_func(model(xb, mask, indices), yb)

    if opt is not None:
        loss.backward()
        if gradient_clipping_norm:
            clip_grad_norm_(model.parameters(), gradient_clipping_norm)
        opt.step()
        opt.zero_grad()

    return loss.item(), len(xb)


def metric_on_batch(metric, model, xb, yb, indices):
    mask = (yb == PADDED_Y_VALUE)
    return metric(model.score(xb, mask, indices), yb)


def metric_on_epoch(metric, model, dl, dev):
    metric_values = torch.mean(
        torch.cat(
            [metric_on_batch(metric, model, xb.to(device=dev), yb.to(device=dev), indices.to(device=dev))
             for xb, yb, indices in dl]
        ), dim=0
    ).cpu().numpy()
    return metric_values


def compute_metrics(metrics, model, dl, dev):
    metric_values_dict = {}
    for metric_name, ats in metrics.items():
        metric_func = getattr(metrics_module, metric_name)
        metric_func_with_ats = partial(metric_func, ats=ats)
        metrics_values = metric_on_epoch(metric_func_with_ats, model, dl, dev)
        metrics_names = ["{metric_name}_{at}".format(metric_name=metric_name, at=at) for at in ats]
        metric_values_dict.update(dict(zip(metrics_names, metrics_values)))

    return metric_values_dict


def epoch_summary(epoch, train_loss, val_loss, train_metrics, val_metrics):
    summary = "Epoch : {epoch} Train loss: {train_loss} Val loss: {val_loss}".format(
        epoch=epoch, train_loss=train_loss, val_loss=val_loss)
    for metric_name, metric_value in train_metrics.items():
        summary += " Train {metric_name} {metric_value}".format(
            metric_name=metric_name, metric_value=metric_value)

    for metric_name, metric_value in val_metrics.items():
        summary += " Val {metric_name} {metric_value}".format(
            metric_name=metric_name, metric_value=metric_value)

    return summary


def get_current_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group["lr"]


def fit(epochs, model, loss_func, optimizer, scheduler, train_dl, valid_dl, config,
        gradient_clipping_norm, early_stopping_patience, device, output_dir, tensorboard_output_path):
    tensorboard_summary_writer = TensorboardSummaryWriter(tensorboard_output_path)

    num_params = get_num_params(model)
    log_num_params(num_params)

    early_stop = EarlyStop(early_stopping_patience)

    for epoch in range(epochs):
        logger.info("Current learning rate: {}".format(get_current_lr(optimizer)))

        model.train()
        # xb dim: [batch_size, slate_length, embedding_dim]
        # yb dim: [batch_size, slate_length]

        train_losses, train_nums = zip(
            *[loss_batch(model, loss_func, xb.to(device=device), yb.to(device=device), indices.to(device=device),
                         gradient_clipping_norm, optimizer) for
              xb, yb, indices in train_dl])
        train_loss = np.sum(np.multiply(train_losses, train_nums)) / np.sum(train_nums)
        train_metrics = compute_metrics(config.metrics, model, train_dl, device)

        model.eval()
        with torch.no_grad():
            val_losses, val_nums = zip(
                *[loss_batch(model, loss_func, xb.to(device=device), yb.to(device=device), indices.to(device=device),
                             gradient_clipping_norm) for
                  xb, yb, indices in valid_dl])
            val_metrics = compute_metrics(config.metrics, model, valid_dl, device)

        val_loss = np.sum(np.multiply(val_losses, val_nums)) / np.sum(val_nums)

        tensorboard_metrics_dict = {("train", "loss"): train_loss, ("val", "loss"): val_loss}

        train_metrics_to_tb = {("train", name): value for name, value in train_metrics.items()}
        tensorboard_metrics_dict.update(train_metrics_to_tb)
        val_metrics_to_tb = {("val", name): value for name, value in val_metrics.items()}
        tensorboard_metrics_dict.update(val_metrics_to_tb)
        tensorboard_metrics_dict.update({("train", "lr"): get_current_lr(optimizer)})

        tensorboard_summary_writer.save_to_tensorboard(tensorboard_metrics_dict, epoch)

        logger.info(epoch_summary(epoch, train_loss, val_loss, train_metrics, val_metrics))

        current_val_metric_value = val_metrics.get(config.val_metric)
        if scheduler:
            if type(scheduler) == torch.optim.lr_scheduler.ReduceLROnPlateau:
                args = [val_metrics[config.val_metric]]
                scheduler.step(*args)
            else:
                scheduler.step()

        early_stop.step(current_val_metric_value, epoch)
        if early_stop.stop_training(epoch):
            logger.info(
                "early stopping at epoch {} since {} didn't improve from epoch no {}. Best value {}, current value {}".format(
                    epoch, config.val_metric, early_stop.best_epoch, early_stop.best_value, current_val_metric_value
                ))
            break

    torch.save(model.state_dict(), os.path.join(output_dir, "model.pkl"))
    tensorboard_summary_writer.close_all_writers()

    return {
        "epochs": epoch,
        "train_metrics": train_metrics,
        "val_metrics": val_metrics,
        "num_params": num_params
    }



import math

import torch
from pytest import approx

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses.lambdaLoss import lambdaLoss


def test_ndcgloss1_ignores_padded():
    y_pred = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, 1.0]])]
    y_true = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, PADDED_Y_VALUE]])]

    result = lambdaLoss(y_pred[0], y_true[0], weighing_scheme="ndcgLoss1_scheme", reduction_log="binary").item()
    result_pad = lambdaLoss(y_pred[1], y_true[1], weighing_scheme="ndcgLoss1_scheme", reduction_log="binary").item()

    expected = 2.9272110462

    assert math.isfinite(result) and math.isfinite(result_pad)
    assert (result == approx(result_pad)) and (result == approx(expected))


def test_ndcgloss2_ignores_padded():
    y_pred = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, 1.0]])]
    y_true = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, PADDED_Y_VALUE]])]

    result = lambdaLoss(y_pred[0], y_true[0], weighing_scheme="ndcgLoss2PP_scheme", reduction_log="binary").item()
    result_pad = lambdaLoss(y_pred[1], y_true[1], weighing_scheme="ndcgLoss2PP_scheme", reduction_log="binary").item()

    expected = 1.1244146823

    assert math.isfinite(result) and math.isfinite(result_pad)
    assert (result == approx(result_pad)) and (result == approx(expected))


def test_ranknet_ignores_padded():
    y_pred = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, 1.0]])]
    y_true = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, PADDED_Y_VALUE]])]

    result = lambdaLoss(y_pred[0], y_true[0], weighing_scheme="rankNet_scheme", reduction_log="natural").item()
    result_pad = lambdaLoss(y_pred[1], y_true[1], weighing_scheme="rankNet_scheme", reduction_log="natural").item()

    expected = 1.1962778568

    assert math.isfinite(result) and math.isfinite(result_pad)
    assert (result == approx(result_pad)) and (result == approx(expected))



from allrank.utils.ltr_logging import get_logger

logger = get_logger()


class EarlyStop:
    def __init__(self, patience):
        self.patience = patience
        self.best_value = 0.0
        self.best_epoch = 0

    def step(self, current_value, current_epoch):
        logger.info("Current:{} Best:{}".format(current_value, self.best_value))
        if current_value > self.best_value:
            self.best_value = current_value
            self.best_epoch = current_epoch

    def stop_training(self, current_epoch) -> bool:
        return current_epoch - self.best_epoch > self.patience



import os
from typing import Any, Dict, Tuple

from tensorboardX import SummaryWriter


class TensorboardSummaryWriter:
    def __init__(self, output_path: str) -> None:
        self.output_path = output_path
        self.writers = {}  # type: Dict[str, Any]

    def ensure_writer_exists(self, name: str) -> None:
        if name not in self.writers.keys():
            writer_path = os.path.join(self.output_path, name)
            self.writers[name] = SummaryWriter(writer_path)

    def save_to_tensorboard(self, results: Dict[Tuple[str, str], float], n_epoch: int) -> None:
        for (role, metric), value in results.items():
            metric_with_role = "_".join([metric, role])
            self.ensure_writer_exists(metric_with_role)
            self.writers[metric_with_role].add_scalar(metric, value, n_epoch)

    def close_all_writers(self) -> None:
        for writer in self.writers.values():
            writer.close()



import torch

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import DEFAULT_EPS


def approxNDCGLoss(y_pred, y_true, eps=DEFAULT_EPS, padded_value_indicator=PADDED_Y_VALUE, alpha=1.):
    """
    Loss based on approximate NDCG introduced in "A General Approximation Framework for Direct Optimization of
    Information Retrieval Measures". Please note that this method does not implement any kind of truncation.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param eps: epsilon value, used for numerical stability
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :param alpha: score difference weight used in the sigmoid function
    :return: loss value, a torch.Tensor
    """
    device = y_pred.device
    y_pred = y_pred.clone()
    y_true = y_true.clone()

    padded_mask = y_true == padded_value_indicator
    y_pred[padded_mask] = float("-inf")
    y_true[padded_mask] = float("-inf")

    # Here we sort the true and predicted relevancy scores.
    y_pred_sorted, indices_pred = y_pred.sort(descending=True, dim=-1)
    y_true_sorted, _ = y_true.sort(descending=True, dim=-1)

    # After sorting, we can mask out the pairs of indices (i, j) containing index of a padded element.
    true_sorted_by_preds = torch.gather(y_true, dim=1, index=indices_pred)
    true_diffs = true_sorted_by_preds[:, :, None] - true_sorted_by_preds[:, None, :]
    padded_pairs_mask = torch.isfinite(true_diffs)
    padded_pairs_mask.diagonal(dim1=-2, dim2=-1).zero_()

    # Here we clamp the -infs to get correct gains and ideal DCGs (maxDCGs)
    true_sorted_by_preds.clamp_(min=0.)
    y_true_sorted.clamp_(min=0.)

    # Here we find the gains, discounts and ideal DCGs per slate.
    pos_idxs = torch.arange(1, y_pred.shape[1] + 1).to(device)
    D = torch.log2(1. + pos_idxs.float())[None, :]
    maxDCGs = torch.sum((torch.pow(2, y_true_sorted) - 1) / D, dim=-1).clamp(min=eps)
    G = (torch.pow(2, true_sorted_by_preds) - 1) / maxDCGs[:, None]

    # Here we approximate the ranking positions according to Eqs 19-20 and later approximate NDCG (Eq 21)
    scores_diffs = (y_pred_sorted[:, :, None] - y_pred_sorted[:, None, :])
    scores_diffs[~padded_pairs_mask] = 0.
    approx_pos = 1. + torch.sum(padded_pairs_mask.float() * (torch.sigmoid(-alpha * scores_diffs).clamp(min=eps)), dim=-1)
    approx_D = torch.log2(1. + approx_pos)
    approx_NDCG = torch.sum((G / approx_D), dim=-1)

    return -torch.mean(approx_NDCG)



import importlib
from typing import List, Any


def instantiate_class(module_name: str, class_name: str):
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_()


class dummy_context_mgr():
    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        return False


def all_equal(values: List[Any]) -> bool:
    return len(set(values)) == 1



import math

import torch
from pytest import approx

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses.approxNDCG import approxNDCGLoss


def test_approxndcg_ignores_padded():
    y_pred = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, 1.0]])]
    y_true = [torch.tensor([[0.5, 0.3, 0.5]]), torch.tensor([[0.5, 0.3, 0.5, PADDED_Y_VALUE]])]

    result = approxNDCGLoss(y_pred[0], y_true[0], alpha=1.).item()
    result_pad = approxNDCGLoss(y_pred[1], y_true[1], alpha=1.).item()

    expected = -0.8499219417

    assert math.isfinite(result) and math.isfinite(result_pad)
    assert (result == approx(result_pad)) and (result == approx(expected))



import json
import os
from argparse import Namespace
from typing import Dict, Any

from attr import asdict
from flatten_dict import flatten

from allrank.config import Config
from allrank.utils.ltr_logging import get_logger

logger = get_logger()


def unpack_numpy_values(dict):
    return {k: v.item() for k, v in dict.items()}


def dump_experiment_result(args: Namespace, config: Config, output_dir: str, result: Dict[str, Any]):
    final_config_dict = asdict(config)
    flattened_experiment = flatten(final_config_dict, reducer="path")
    result["train_metrics"] = unpack_numpy_values(result["train_metrics"])
    result["val_metrics"] = unpack_numpy_values(result["val_metrics"])
    result["num_params"] = result["num_params"].item()
    flattened_result = flatten(result, reducer="path")
    flattened_experiment.update(flattened_result)
    flattened_experiment["run_id"] = args.run_id
    flattened_experiment["dir"] = output_dir
    with open(os.path.join(output_dir, "experiment_result.json"), "w") as json_file:
        json.dump(flattened_experiment, json_file)
        json_file.write("\n")


def assert_expected_metrics(result: Dict[str, Any], expected_metrics: Dict[str, Dict[str, float]]):
    if expected_metrics:
        for role, metrics in expected_metrics.items():
            for name, expected_value in metrics.items():
                actual_value = result["{}_metrics".format(role)][name]
                msg = "{} {} got {}. It was expected to be at least {}".format(
                    role, name, actual_value, expected_value)
                if actual_value < expected_value:
                    logger.info(msg)
                assert actual_value >= expected_value, msg



import os

from allrank.utils.ltr_logging import get_logger

logger = get_logger()


def execute_command(command):
    logger.info("will execute {}".format(command))
    result = os.system(command)
    logger.info("exit_code = {}".format(result))
    if result != 0:
        raise RuntimeError("non-zero exit-code: {} from command '{}'".format(result, command))



import math

import numpy as np
import torch
from pytest import approx
from scipy.special import softmax

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import DEFAULT_EPS, binary_listNet


def binary_listNet_wrap(y_pred, y_true, eps=1e-10):
    return binary_listNet(torch.tensor([y_pred]), torch.tensor([y_true]), eps).item()


def test_binary_listnet_simple():
    y_pred = [0.5, 0.2]
    y_true = [1.0, 0.0]

    result = binary_listNet_wrap(y_pred, y_true, eps=0.0)
    expected = - np.sum(y_true * np.log(softmax(y_pred)))

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_binary_listnet_stable_for_very_small_prediction():
    y_pred = [0.5, -1e30]
    y_true = [1.0, 0.0]

    result = binary_listNet_wrap(y_pred, y_true)
    expected = - np.sum(y_true * np.log(softmax(y_pred) + DEFAULT_EPS))

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected, abs=1e-9))


def test_binary_listnet_ignores_padded_value():
    y_pred = [0.5, 0.2, 0.5]
    y_true = [1.0, 0.0, PADDED_Y_VALUE]

    result = binary_listNet_wrap(y_pred, y_true)
    expected = - np.sum(y_true[:2] * np.log(softmax(y_pred[:2]) + DEFAULT_EPS))

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))



from typing import Any

import numpy as np
import torch
import torch.nn as nn

from allrank.utils.file_utils import is_gs_path, copy_file_to_local
from allrank.utils.ltr_logging import get_logger

logger = get_logger()


def get_torch_device():
    """
    Getter for an available pyTorch device.
    :return: CUDA-capable GPU if available, CPU otherwise
    """
    return torch.device("cuda:0") if torch.cuda.is_available() else torch.device("cpu")


def get_num_params(model: nn.Module) -> int:
    """
    Calculation of the number of nn.Module parameters.
    :param model: nn.Module
    :return: number of parameters
    """
    model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    params = sum([np.prod(p.size()) for p in model_parameters])
    return params  # type: ignore


def log_num_params(num_params: int) -> None:
    """
    Logging num_params to the global logger.
    :param num_params: number of parameters to log
    """
    logger.info("Model has {} trainable parameters".format(num_params))


class CustomDataParallel(nn.DataParallel):
    """
    Wrapper for scoring with nn.DataParallel object containing LTRModel.
    """

    def score(self, x, mask, indices):
        """
        Wrapper function for a forward pass through the whole LTRModel and item scoring.
        :param x: input of shape [batch_size, slate_length, input_dim]
        :param mask: padding mask of shape [batch_size, slate_length]
        :param indices: original item ranks used in positional encoding, shape [batch_size, slate_length]
        :return: scores of shape [batch_size, slate_length]
        """
        return self.module.score(x, mask, indices)  # type: ignore


def load_state_dict_from_file(path: str, device: Any):
    if is_gs_path(path):
        path = copy_file_to_local(path)

    return torch.load(path, map_location=device)



import numpy as np

from allrank.click_models.base import FixedClickModel, MultipleClickModel, ConditionedClickModel
from tests.click_models import click


def test_click_model_should_use_all_click_models():
    np.random.seed(42)

    click_model_0 = FixedClickModel([0])
    click_model_1 = FixedClickModel([1])
    click_model = MultipleClickModel([click_model_0, click_model_1], probabilities=[0.5, 0.5])
    clicks = np.array([
        click(click_model, [], [1, 2])
        for _ in range(20000)
    ])
    assert 9950 < np.sum(clicks[:, 0]) < 10050
    assert 9950 < np.sum(clicks[:, 1]) < 10050


def test_click_model_should_combine_click_models_and():
    click_model_0 = FixedClickModel([0, 1])
    click_model_1 = FixedClickModel([1, 2])
    click_model = ConditionedClickModel([click_model_0, click_model_1], np.all)
    clicks = click(click_model, [], [1, 2, 3])
    assert clicks == [0, 1, 0]


def test_click_model_should_combine_click_models_or():
    click_model_0 = FixedClickModel([0, 1])
    click_model_1 = FixedClickModel([1, 2])
    click_model = ConditionedClickModel([click_model_0, click_model_1], np.any)
    clicks = click(click_model, [], [1, 2, 3, 4])
    assert clicks == [1, 1, 1, 0]



import numpy as np

from allrank.click_models.base import RandomClickModel
from tests.click_models import click


def test_random_click_model_single():
    click_model = RandomClickModel(1)
    np.random.seed(42)
    assert click(click_model, [], [1]) == [1]
    assert click(click_model, [], [1, 2]) == [0, 1]
    assert click(click_model, [], [1, 2, 3]) == [0, 1, 0]


def test_random_click_model_multiple():
    np.random.seed(42)
    assert click(RandomClickModel(2), [], [1, 2, 3, 4]) == [0, 1, 0, 1]
    assert click(RandomClickModel(3), [], [1, 2, 3, 4]) == [1, 1, 0, 1]
    assert click(RandomClickModel(4), [], [1, 2, 3, 4]) == [1, 1, 1, 1]



import torch
import torch.nn.functional as F

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import DEFAULT_EPS


def listNet(y_pred, y_true, eps=DEFAULT_EPS, padded_value_indicator=PADDED_Y_VALUE):
    """
    ListNet loss introduced in "Learning to Rank: From Pairwise Approach to Listwise Approach".
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param eps: epsilon value, used for numerical stability
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: loss value, a torch.Tensor
    """
    y_pred = y_pred.clone()
    y_true = y_true.clone()

    mask = y_true == padded_value_indicator
    y_pred[mask] = float('-inf')
    y_true[mask] = float('-inf')

    preds_smax = F.softmax(y_pred, dim=1)
    true_smax = F.softmax(y_true, dim=1)

    preds_smax = preds_smax + eps
    preds_log = torch.log(preds_smax)

    return torch.mean(-torch.sum(true_smax * preds_log, dim=1))



import numpy as np

from allrank.click_models.cascade_models import BaseCascadeModel, DiverseClicksModel
from tests.click_models import click

base_click_model = BaseCascadeModel(0.0, 1)
click_model = DiverseClicksModel(base_click_model)


def test_diverse_clicks_model_simple():
    assert click(click_model, np.array([[0, 1]]), [1]) == [1]
    assert click(click_model, np.array([[0, 1], [0, 1]]), [1, 1]) == [1, 0]
    assert click(click_model, np.array([[0, 1], [0, 1], [1, 1]]), [1, 1, 1]) == [1, 0, 0]
    assert click(click_model, np.array([[0, 1], [0, 1], [2, 2], [1, 1]]), [1, 1, 1, 1]) == [1, 0, 1, 0]



"""
Code in this file was adapted from "The Annotated Transformer" by Harvard NLP.
http://nlp.seas.harvard.edu/2018/04/03/attention.html
"""

import math
from typing import Optional

import torch
import torch.nn as nn

from allrank.config import PositionalEncoding


class FixedPositionalEncoding(nn.Module):
    """
    Class implementing fixed positional encodings.

    Fixed positional encodings up to max_len position are computed once during object construction.
    """
    def __init__(self, d_model: int, max_len=5000):
        """
        :param d_model: dimensionality of the embeddings
        :param max_len: maximum length of the sequence
        """
        super().__init__()

        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0.0, max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0.0, d_model, 2) *
                             -(math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = torch.cat((pe, torch.zeros([1, d_model])))
        self.padding_idx = pe.size()[0] - 1
        self.register_buffer('pe', pe)

    def forward(self, x, mask, indices):
        """
        Forward pass through the FixedPositionalEncoding.
        :param x: input of shape [batch_size, slate_length, d_model]
        :param mask: padding mask of shape [batch_size, slate_length]
        :param indices: original item ranks used in positional encoding, shape [batch_size, slate_length]
        :return: output of shape [batch_size, slate_length, d_model]
        """
        padded_indices = indices.masked_fill(mask, self.padding_idx)
        padded_indices[padded_indices > self.padding_idx] = self.padding_idx
        x = math.sqrt(self.pe.shape[1]) * x + self.pe[padded_indices, :]  # type: ignore
        return x


class LearnedPositionalEncoding(nn.Module):
    """
    Class implementing learnable positional encodings.
    """
    def __init__(self, d_model, max_len=5000):
        """
        :param d_model: dimensionality of the embeddings
        :param max_len: maximum length of the sequence
        """
        super().__init__()

        self.pe = nn.Embedding(max_len + 1, d_model, padding_idx=-1)

    def forward(self, x, mask, indices):
        """
        Forward pass through the LearnedPositionalEncoding.
        :param x: input of shape [batch_size, slate_length, d_model]
        :param mask: padding mask of shape [batch_size, slate_length]
        :param indices: original item ranks used in positional encoding, shape [batch_size, slate_length]
        :return: output of shape [batch_size, slate_length, d_model]
        """
        padded_indices = indices.masked_fill(mask, self.pe.padding_idx)
        padded_indices[padded_indices > self.pe.padding_idx] = self.pe.padding_idx
        x = math.sqrt(self.pe.embedding_dim) * x + self.pe(padded_indices)
        return x


def _make_positional_encoding(d_model: int, positional_encoding: Optional[PositionalEncoding]):
    """
    Helper function for instantiating positional encodings classes.
    :param d_model: dimensionality of the embeddings
    :param positional_encoding: config.PositionalEncoding object containing PE config
    :return: positional encoding object of given variant
    """
    if positional_encoding is None:
        return None
    elif positional_encoding.strategy == "fixed":
        return FixedPositionalEncoding(d_model, max_len=positional_encoding.max_indices)
    elif positional_encoding.strategy == "learned":
        return LearnedPositionalEncoding(d_model, max_len=positional_encoding.max_indices)
    else:
        raise ValueError("Invalid positional encoding type: {}".format(positional_encoding.strategy))



import math

import numpy as np
import torch
from pytest import approx

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import (
    ordinal,
    with_ordinals)


def loss_wrap(y_pred, y_true):
    return ordinal(torch.tensor([y_pred]), torch.tensor([y_true]), n=2).item()


def xe(true, pred):
    return - true * math.log(pred) - (1 - true) * math.log(1 - pred)


def test_ds_transform():
    y_true = np.array([2.0, 1.0, 0.0])
    result = with_ordinals(torch.tensor([y_true], dtype=torch.float), 2).tolist()
    expected = [[[1.0, 1.0], [1.0, 0.0], [0.0, 0.0]]]
    assert result == expected


def test_ordinal_single_doc():
    y_pred = [[0.8, 0.6]]
    y_true = [1.0]

    result = loss_wrap(y_pred, y_true)
    expected = np.mean([xe(1, 0.8) + xe(0, 0.6)])

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ordinal_simple():
    y_pred = [[0.8, 0.7], [0.4, 0.3], [0.2, 0.1]]
    y_true = [2.0, 1.0, 0.0]

    result = loss_wrap(y_pred, y_true)
    expected = np.mean([xe(1, 0.8) + xe(1, 0.7), xe(1, 0.4) + xe(0, 0.3), xe(0, 0.2) + xe(0, 0.1)])

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ordinal_single_doc_padded():
    y_pred = [[0.8, 0.6], [0.2, 0.1]]
    y_true = [1.0, PADDED_Y_VALUE]

    result = loss_wrap(y_pred, y_true)
    expected = np.mean([xe(1, 0.8) + xe(0, 0.6)])

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))



import torch

from allrank.models.losses.neuralNDCG import neuralNDCG, neuralNDCG_transposed
from allrank.models.metrics import ndcg


def neuralNDCG_wrap(
        y_pred, y_true, temperature=1e-4, powered_relevancies=True, k=None, stochastic=False,
        n_samples=1024, beta=0.001, transposed=False):
    if transposed:
        fun = neuralNDCG_transposed  # type: ignore
    else:
        fun = neuralNDCG  # type: ignore

    return fun(torch.tensor([y_pred]), torch.tensor([y_true]), temperature=temperature,
               powered_relevancies=powered_relevancies, k=k,
               stochastic=stochastic, n_samples=n_samples, beta=beta).item()


def ndcg_wrap(y_pred, y_true, ats=None):
    return ndcg(torch.tensor([y_pred]), torch.tensor([y_true]), ats=ats).mean().item()



import math

import torch
from pytest import approx

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.metrics import ndcg


def ndcg_wrap(y_pred, y_true, ats=None):
    return ndcg(torch.tensor([y_pred]), torch.tensor([y_true]), ats=ats).numpy()


def test_ndcg_simple_1():
    y_pred = [0.5, 0.2]
    y_true = [1.0, 0.0]

    result = ndcg_wrap(y_pred, y_true)

    assert (result == 1.0)


def test_ndcg_simple_2():
    y_pred = [0.5, 0.2]
    y_true = [0.0, 1.0]

    result = ndcg_wrap(y_pred, y_true)

    assert (result == 1 / math.log2(3))


def test_ndcg_zero_when_no_relevant():
    y_pred = [0.5, 0.2]
    y_true = [0.0, 0.0]

    result = ndcg_wrap(y_pred, y_true)

    assert (result == 0.0)


def test_ndcg_for_multiple_ats():
    y_pred = [0.5, 0.2, 0.1]
    y_true = [1.0, 0.0, 1.0]

    result = ndcg_wrap(y_pred, y_true, ats=[1, 2])

    ndcg_one_relevant_on_top = 1.0 / (1.0 + 1 / math.log2(3))
    expected = [1.0, ndcg_one_relevant_on_top]

    batch_0 = 0
    assert result[batch_0] == approx(expected)


def test_ndcg_with_padded_input():
    y_pred = [0.5, 0.2, 1.0]
    y_true = [1.0, 0.0, PADDED_Y_VALUE]

    result = ndcg_wrap(y_pred, y_true)

    assert result == 1.0


def test_ndcg_with_padded_input_2():
    y_pred = [0.5, 0.2, 1.0]
    y_true = [0.0, 1.0, PADDED_Y_VALUE]

    result = ndcg_wrap(y_pred, y_true)

    assert result == 1 / math.log2(3)



import math
from abc import ABC, abstractmethod
from typing import List, Tuple, Callable

import numpy as np
import torch


class ClickModel(ABC):
    """
    Base class for all click models. Specifies the click model contract
    """

    @abstractmethod
    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        """
        Applies a click model and returns the mask for documents.

        :rtype: np.ndarray [ number_of_documents ] -> a mask of the same length as the documents -
        defining whether a document was clicked (1), not clicked (0) or is a padded element (-1)

        :param documents: Tuple of :
           torch.Tensor [ number_of_documents, dimensionality_of_latent_vector ], representing features of documents
           torch.Tensor [ number_of_documents ] representing relevancy of documents
        """
        pass


class RandomClickModel(ClickModel):
    """
    This ClickModel clicks a configured number of times on random documents
    """

    def __init__(self, n_clicks: int):
        """

        :param n_clicks: number of documents that will be clicked
        """
        self.n_clicks = n_clicks

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        X, y = documents
        clicks = np.random.choice(range(len(y)), size=self.n_clicks, replace=False)
        mask = np.zeros(len(y), dtype=bool)
        mask[clicks] = 1
        return mask


class FixedClickModel(ClickModel):
    """
    This ClickModel clicks on documents at fixed positions
    """

    def __init__(self, click_positions: List[int]):
        """

        :param click_positions: list of indices of documents that will be clicked
        """
        self.click_positions = click_positions

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        X, y = documents
        clicks = np.zeros(len(y), dtype=bool)
        clicks[self.click_positions] = 1
        return clicks


class MultipleClickModel(ClickModel):
    """
    This click model uses one of given click models with given probability
    """

    def __init__(self, inner_click_models: List[ClickModel], probabilities: List[float]):
        """

        :param inner_click_models: list of click models to choose from
        :param probabilities: list of probabilities - must be of the same length as list of click models and sum to 1.0
        """
        self.inner_click_models = inner_click_models
        assert math.isclose(np.sum(probabilities), 1.0, abs_tol=1e-5), \
            f"probabilities should sum to one, but got {probabilities} which sums to {np.sum(probabilities)}"
        self.probabilities = np.array(probabilities).cumsum()

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        index = np.argmax(np.random.rand() < self.probabilities)
        result = self.inner_click_models[index].click(documents)  # type: ignore
        return result


class ConditionedClickModel(ClickModel):
    """
    This click model allows to combine multiple click models with a logical funciton
    """

    def __init__(self, inner_click_models: List[ClickModel], combiner: Callable):
        """

        :param inner_click_models: list of click models to combine
        :param combiner: a function applied to the result of clicks from click models - e.g. np.all or np.any
        """
        self.inner_click_models = inner_click_models
        self.combiner = combiner

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        clicks_from_click_models = [click_model.click(documents) for click_model in self.inner_click_models]
        return self.combiner(clicks_from_click_models, 0)


class MaxClicksModel(ClickModel):
    """
    This click model takes other click model and limits the number of clicks to given value
    effectively keeping top `max_clicks` clicks
    """

    def __init__(self, inner_click_model: ClickModel, max_clicks: int):
        """

        :param inner_click_model: a click model to generate clicks
        :param max_clicks: number of clicks that should be preserved
        """
        self.inner_click_model = inner_click_model
        self.max_clicks = max_clicks

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        underlying_clicks = self.inner_click_model.click(documents)
        if self.max_clicks is not None:
            max_clicks_mask = underlying_clicks.cumsum() <= self.max_clicks
            return underlying_clicks * max_clicks_mask
        return underlying_clicks


class OnlyRelevantClickModel(ClickModel):
    """
    This ClickModel clicks on a document when its relevancy is greater that or equal to a predefined value

    """

    def __init__(self, relevancy_threshold: float):
        """
        :param relevancy_threshold: a minimum value of relevancy of a document to be clicked (inclusive)
        """
        self.relevancy_threshold = relevancy_threshold

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        X, y = documents
        return np.array(y) >= self.relevancy_threshold



from typing import Iterable

import numpy as np
from sklearn.datasets import dump_svmlight_file

from allrank.data.dataset_loading import PADDED_Y_VALUE


def write_to_libsvm_without_masked(path: str, X: Iterable[np.ndarray], y: Iterable[np.ndarray]) -> None:
    """
    This function writes given X's and y's in svmlight / libsvm file format.
    It supports padded documents - they are removed from the written dataset.
    Slates are identified by a 'qid' column within the file.

    :param path: a path to save libsvm file
    :param X: Iterable of lists of document vectors
    :param y: Iterable of lists of document relevancies
    """
    Xs = []
    ys = []
    qids = []
    qid = 0
    for X, y in zip(X, y):
        mask = y != PADDED_Y_VALUE
        Xs.append(X[mask])  # type:ignore
        ys.append(y[mask])  # type:ignore
        qids.append(np.repeat(qid, len(y[mask])))  # type:ignore
        qid += 1
    Xs = np.vstack(Xs)  # type:ignore
    ys = np.concatenate(ys)
    qids = np.concatenate(qids)
    dump_svmlight_file(Xs, ys, path, query_id=qids)



DEFAULT_EPS = 1e-10

from .approxNDCG import *  # noqa F403 F401
from .binary_listNet import *  # noqa F403 F401
from .lambdaLoss import *  # noqa F403 F401
from .listMLE import *  # noqa F403 F401
from .listNet import *  # noqa F403 F401
from .neuralNDCG import *  # noqa F403 F401
from .ordinal import *  # noqa F403 F401
from .pointwise import *  # noqa F403 F401
from .rankNet import *  # noqa F403 F401
from .bce import *  # noqa F403 F401



from typing import Tuple

import numpy as np
import torch
from scipy.spatial.distance import cdist

from allrank.click_models.base import ClickModel
from allrank.click_models.duplicate_aware import EverythingButDuplicatesClickModel
from allrank.data.dataset_loading import PADDED_Y_VALUE


class BaseCascadeModel(ClickModel):
    """
    This ClickModel simulates decaying probability of observing an item
    and clicks on an observed item given it's relevance is greater than or equal to a given threshold

    """

    def __init__(self, eta: float, threshold: float):
        """

        :param eta: the power to be applied over a result of a decay function (specified as 1/position)
                    to decide whether a document was observed
        :param threshold: a minimum value of relevancy of an observed document to be clicked (inclusive)
        """
        self.eta = eta
        self.threshold = threshold

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        X, y = documents
        observed_mask = (1 / np.arange(1, len(y) + 1) ** self.eta) >= np.random.rand(len(y))
        return (y * observed_mask >= self.threshold).numpy()


class DiverseClicksModel(ClickModel):
    """
    A 'diverse-clicks' model from Seq2Slate paper https://arxiv.org/abs/1810.02019
    It clicks on documents from top to the bottom if:
      1. a delegate click model decides to click on the document (in the original paper - CascadeModel
      2. it is no closer than a defined percentile of distances to a previously clicked document
    """

    def __init__(self, inner_click_model, q_percentile=0.5):
        """

        :param inner_click_model: original, non-diversified click model
        :param q_percentile: a percentile of pairwise distances that will be used as a distance threshold to tell if a pair is a duplicate
        """
        self.inner_click_model = inner_click_model
        self.q_percentile = q_percentile

    def __pairwise_distances_list(self, X):
        dist = cdist(X, X, metric='euclidean')
        triu_indices = np.triu_indices(dist.shape[0] - 1)
        return dist[:-1, 1:][triu_indices]

    def click(self, documents: Tuple[torch.Tensor, torch.Tensor]) -> np.ndarray:
        X, y = documents

        real_docs_mask = (y != PADDED_Y_VALUE)
        real_X = X[real_docs_mask, :]

        distances = self.__pairwise_distances_list(real_X)
        if len(distances) == 0:
            duplicate_margin = 0
        else:
            duplicate_margin = np.quantile(distances, q=self.q_percentile)

        def not_similar(x_vec, clicked_X):
            cX = clicked_X.copy()
            cX.append(x_vec)
            cX = torch.stack(cX, dim=0)
            cm = EverythingButDuplicatesClickModel(duplicate_margin)
            clicks = cm.click((cX, np.ones(len(cX))))
            last_element_clicked = clicks[-1]
            return last_element_clicked == 1

        relevant_for_click = self.inner_click_model.click(documents)

        clicked_Xs = []  # type: ignore
        indices_to_click = np.argwhere(relevant_for_click == 1)
        for idx_to_click in indices_to_click:
            idx_to_click = idx_to_click[0]
            X_to_click = X[idx_to_click]
            if not_similar(X_to_click, clicked_Xs):
                clicked_Xs.append(X_to_click)
            else:
                relevant_for_click[idx_to_click] = 0

        return relevant_for_click



import torch

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import DEFAULT_EPS
from allrank.models.losses.loss_utils import deterministic_neural_sort, sinkhorn_scaling, stochastic_neural_sort
from allrank.models.metrics import dcg
from allrank.models.model_utils import get_torch_device


def neuralNDCG(y_pred, y_true, padded_value_indicator=PADDED_Y_VALUE, temperature=1., powered_relevancies=True, k=None,
               stochastic=False, n_samples=32, beta=0.1, log_scores=True):
    """
    NeuralNDCG loss introduced in "NeuralNDCG: Direct Optimisation of a Ranking Metric via Differentiable
    Relaxation of Sorting" - https://arxiv.org/abs/2102.07831. Based on the NeuralSort algorithm.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :param temperature: temperature for the NeuralSort algorithm
    :param powered_relevancies: whether to apply 2^x - 1 gain function, x otherwise
    :param k: rank at which the loss is truncated
    :param stochastic: whether to calculate the stochastic variant
    :param n_samples: how many stochastic samples are taken, used if stochastic == True
    :param beta: beta parameter for NeuralSort algorithm, used if stochastic == True
    :param log_scores: log_scores parameter for NeuralSort algorithm, used if stochastic == True
    :return: loss value, a torch.Tensor
    """
    dev = get_torch_device()

    if k is None:
        k = y_true.shape[1]

    mask = (y_true == padded_value_indicator)
    # Choose the deterministic/stochastic variant
    if stochastic:
        P_hat = stochastic_neural_sort(y_pred.unsqueeze(-1), n_samples=n_samples, tau=temperature, mask=mask,
                                       beta=beta, log_scores=log_scores)
    else:
        P_hat = deterministic_neural_sort(y_pred.unsqueeze(-1), tau=temperature, mask=mask).unsqueeze(0)

    # Perform sinkhorn scaling to obtain doubly stochastic permutation matrices
    P_hat = sinkhorn_scaling(P_hat.view(P_hat.shape[0] * P_hat.shape[1], P_hat.shape[2], P_hat.shape[3]),
                             mask.repeat_interleave(P_hat.shape[0], dim=0), tol=1e-6, max_iter=50)
    P_hat = P_hat.view(int(P_hat.shape[0] / y_pred.shape[0]), y_pred.shape[0], P_hat.shape[1], P_hat.shape[2])

    # Mask P_hat and apply to true labels, ie approximately sort them
    P_hat = P_hat.masked_fill(mask[None, :, :, None] | mask[None, :, None, :], 0.)
    y_true_masked = y_true.masked_fill(mask, 0.).unsqueeze(-1).unsqueeze(0)
    if powered_relevancies:
        y_true_masked = torch.pow(2., y_true_masked) - 1.

    ground_truth = torch.matmul(P_hat, y_true_masked).squeeze(-1)
    discounts = (torch.tensor(1.) / torch.log2(torch.arange(y_true.shape[-1], dtype=torch.float) + 2.)).to(dev)
    discounted_gains = ground_truth * discounts

    if powered_relevancies:
        idcg = dcg(y_true, y_true, ats=[k]).permute(1, 0)
    else:
        idcg = dcg(y_true, y_true, ats=[k], gain_function=lambda x: x).permute(1, 0)

    discounted_gains = discounted_gains[:, :, :k]
    ndcg = discounted_gains.sum(dim=-1) / (idcg + DEFAULT_EPS)
    idcg_mask = idcg == 0.
    ndcg = ndcg.masked_fill(idcg_mask.repeat(ndcg.shape[0], 1), 0.)

    assert (ndcg < 0.).sum() >= 0, "every ndcg should be non-negative"
    if idcg_mask.all():
        return torch.tensor(0.)

    mean_ndcg = ndcg.sum() / ((~idcg_mask).sum() * ndcg.shape[0])  # type: ignore
    return -1. * mean_ndcg  # -1 cause we want to maximize NDCG


def neuralNDCG_transposed(y_pred, y_true, padded_value_indicator=PADDED_Y_VALUE, temperature=1.,
                          powered_relevancies=True, k=None, stochastic=False, n_samples=32, beta=0.1, log_scores=True,
                          max_iter=50, tol=1e-6):
    """
    NeuralNDCG Transposed loss introduced in "NeuralNDCG: Direct Optimisation of a Ranking Metric via Differentiable
    Relaxation of Sorting" - https://arxiv.org/abs/2102.07831. Based on the NeuralSort algorithm.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :param temperature: temperature for the NeuralSort algorithm
    :param powered_relevancies: whether to apply 2^x - 1 gain function, x otherwise
    :param k: rank at which the loss is truncated
    :param stochastic: whether to calculate the stochastic variant
    :param n_samples: how many stochastic samples are taken, used if stochastic == True
    :param beta: beta parameter for NeuralSort algorithm, used if stochastic == True
    :param log_scores: log_scores parameter for NeuralSort algorithm, used if stochastic == True
    :param max_iter: maximum iteration count for Sinkhorn scaling
    :param tol: tolerance for Sinkhorn scaling
    :return: loss value, a torch.Tensor
    """
    dev = get_torch_device()

    if k is None:
        k = y_true.shape[1]

    mask = (y_true == padded_value_indicator)

    if stochastic:
        P_hat = stochastic_neural_sort(y_pred.unsqueeze(-1), n_samples=n_samples, tau=temperature, mask=mask,
                                       beta=beta, log_scores=log_scores)
    else:
        P_hat = deterministic_neural_sort(y_pred.unsqueeze(-1), tau=temperature, mask=mask).unsqueeze(0)

    # Perform sinkhorn scaling to obtain doubly stochastic permutation matrices
    P_hat_masked = sinkhorn_scaling(P_hat.view(P_hat.shape[0] * y_pred.shape[0], y_pred.shape[1], y_pred.shape[1]),
                                    mask.repeat_interleave(P_hat.shape[0], dim=0), tol=tol, max_iter=max_iter)
    P_hat_masked = P_hat_masked.view(P_hat.shape[0], y_pred.shape[0], y_pred.shape[1], y_pred.shape[1])
    discounts = (torch.tensor(1) / torch.log2(torch.arange(y_true.shape[-1], dtype=torch.float) + 2.)).to(dev)

    # This takes care of the @k metric truncation - if something is @>k, it is useless and gets 0.0 discount
    discounts[k:] = 0.
    discounts = discounts[None, None, :, None]

    # Here the discounts become expected discounts
    discounts = torch.matmul(P_hat_masked.permute(0, 1, 3, 2), discounts).squeeze(-1)
    if powered_relevancies:
        gains = torch.pow(2., y_true) - 1
        discounted_gains = gains.unsqueeze(0) * discounts
        idcg = dcg(y_true, y_true, ats=[k]).squeeze()
    else:
        gains = y_true
        discounted_gains = gains.unsqueeze(0) * discounts
        idcg = dcg(y_true, y_true, ats=[k]).squeeze()

    ndcg = discounted_gains.sum(dim=2) / (idcg + DEFAULT_EPS)
    idcg_mask = idcg == 0.
    ndcg = ndcg.masked_fill(idcg_mask, 0.)

    assert (ndcg < 0.).sum() >= 0, "every ndcg should be non-negative"
    if idcg_mask.all():
        return torch.tensor(0.)

    mean_ndcg = ndcg.sum() / ((~idcg_mask).sum() * ndcg.shape[0])  # type: ignore
    return -1. * mean_ndcg  # -1 cause we want to maximize NDCG



import math

import torch
from pytest import approx
from torch.nn import BCEWithLogitsLoss

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import rankNet, rankNet_weightByGTDiff, rankNet_weightByGTDiff_pow


def rankNet_wrap(y_pred, y_true):
    return rankNet(torch.tensor([y_pred]), torch.tensor([y_true])).item()


def rankNet_weighted_wrap(y_pred, y_true):
    return rankNet_weightByGTDiff(torch.tensor([y_pred]), torch.tensor([y_true])).item()


def rankNet_weighted_pow_wrap(y_pred, y_true):
    return rankNet_weightByGTDiff_pow(torch.tensor([y_pred]), torch.tensor([y_true])).item()


def bce_wrap(y_pred, y_true, weight=None):
    if weight:
        weight = torch.tensor(weight)
    return BCEWithLogitsLoss(weight=weight)(torch.tensor([y_pred]), torch.tensor([y_true])).item()


def test_ranknet_onepair():
    y_pred = [0.5, 0.2]
    y_true = [1.0, 0.0]

    result = rankNet_wrap(y_pred, y_true)
    expected = bce_wrap(y_pred[0] - y_pred[1], 1.0)

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ranknet_onepair_minus():
    y_pred = [0.2, 0.5]
    y_true = [1.0, 0.0]

    result = rankNet_wrap(y_pred, y_true)
    expected = bce_wrap(y_pred[0] - y_pred[1], 1.0)

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ranknet_two_pairs():
    y_pred = [0.5, 0.2, 0.1]
    y_true = [1.0, 0.0, 0.0]

    result = rankNet_wrap(y_pred, y_true)
    expected = bce_wrap([0.3, 0.4], [1.0, 1.0])

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ranknet_pair_reversed():
    y_pred = [0.2, 0.5]
    y_true = [0.0, 1.0]

    result = rankNet_wrap(y_pred, y_true)
    expected = bce_wrap(0.3, 1.0)

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ranknet_pair_multirelevancy():
    y_pred = [0.2, 0.5]
    y_true = [0.0, 2.0]

    result = rankNet_wrap(y_pred, y_true)
    expected = bce_wrap(0.3, 1.0)

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ranknet_onepair_masked():
    y_pred = [0.5, 0.2, 0.66]
    y_true = [1.0, 0.0, PADDED_Y_VALUE]

    result = rankNet_wrap(y_pred, y_true)
    expected = bce_wrap(y_pred[0] - y_pred[1], 1.0)

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ranknet_two_pairs_weighted():
    y_pred = [0.5, 0.2, 0.1]
    y_true = [2.0, 1.0, 0.0]

    result = rankNet_weighted_wrap(y_pred, y_true)
    expected = bce_wrap([0.3, 0.4, 0.1], [1.0, 1.0, 1.0], weight=[1.0, 2.0, 1.0])

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_ranknet_two_pairs_weighted_pow():
    y_pred = [0.5, 0.2, 0.1]
    y_true = [2.0, 1.0, 0.0]

    result = rankNet_weighted_pow_wrap(y_pred, y_true)
    expected = bce_wrap([0.3, 0.4, 0.1], [1.0, 1.0, 1.0], weight=[3.0, 4.0, 1.0])

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))



import os
from typing import Tuple

import numpy as np
import torch
from sklearn.datasets import load_svmlight_file
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms
from torchvision.transforms import Compose

from allrank.utils.file_utils import open_local_or_gs
from allrank.utils.ltr_logging import get_logger

logger = get_logger()
PADDED_Y_VALUE = -1
PADDED_INDEX_VALUE = -1


class ToTensor(object):
    """
    Wrapper for ndarray->Tensor conversion.
    """
    def __call__(self, sample):
        """
        :param sample: tuple of three ndarrays
        :return: ndarrays converted to tensors
        """
        x, y, indices = sample
        return torch.from_numpy(x).type(torch.float32), torch.from_numpy(y).type(torch.float32), torch.from_numpy(indices).type(torch.long)


class FixLength(object):
    """
    Wrapper for slate transformation to fix its length, either by zero padding or sampling.

    For a given slate, if its length is less than self.dim_given, x's and y's are padded with zeros to match that length.
    If its length is greater than self.dim_given, a random sample of items from that slate is taken to match the self.dim_given.
    """
    def __init__(self, dim_given):
        """
        :param dim_given: dimensionality of x after length fixing operation
        """
        assert isinstance(dim_given, int)
        self.dim_given = dim_given

    def __call__(self, sample):
        """
        :param sample: ndarrays tuple containing features, labels and original ranks of shapes
        [sample_length, features_dim], [sample_length] and [sample_length], respectively
        :return: ndarrays tuple containing features, labels and original ranks of shapes
            [self.dim_given, features_dim], [self.dim_given] and [self.dim_given], respectively
        """
        sample_size = len(sample[1])
        if sample_size < self.dim_given:  # when expected dimension is larger than number of observation in instance do the padding
            fixed_len_x, fixed_len_y, indices = self._pad(sample, sample_size)
        else:  # otherwise do the sampling
            fixed_len_x, fixed_len_y, indices = self._sample(sample, sample_size)

        return fixed_len_x, fixed_len_y, indices

    def _sample(self, sample, sample_size):
        """
        Sampling from a slate longer than self.dim_given.
        :param sample: ndarrays tuple containing features, labels and original ranks of shapes
            [sample_length, features_dim], [sample_length] and [sample_length], respectively
        :param sample_size: target slate length
        :return: ndarrays tuple containing features, labels and original ranks of shapes
            [sample_size, features_dim], [sample_size] and [sample_size]
        """
        indices = np.random.choice(sample_size, self.dim_given, replace=False)
        fixed_len_y = sample[1][indices]
        if fixed_len_y.sum() == 0:
            if sample[1].sum() == 1:
                indices = np.concatenate([np.random.choice(indices, self.dim_given - 1, replace=False), [np.argmax(sample[1])]])
                fixed_len_y = sample[1][indices]
            elif sample[1].sum() > 0:
                return self._sample(sample, sample_size)
        fixed_len_x = sample[0][indices]
        return fixed_len_x, fixed_len_y, indices

    def _pad(self, sample, sample_size):
        """
        Zero padding a slate shorter than self.dim_given
        :param sample: ndarrays tuple containing features, labels and original ranks of shapes
            [sample_length, features_dim], [sample_length] and [sample_length]
        :param sample_size: target slate length
        :return: ndarrays tuple containing features, labels and original ranks of shapes
            [sample_size, features_dim], [sample_size] and [sample_size]
        """
        fixed_len_x = np.pad(sample[0], ((0, self.dim_given - sample_size), (0, 0)), "constant")
        fixed_len_y = np.pad(sample[1], (0, self.dim_given - sample_size), "constant", constant_values=PADDED_Y_VALUE)
        indices = np.pad(np.arange(0, sample_size), (0, self.dim_given - sample_size), "constant", constant_values=PADDED_INDEX_VALUE)
        return fixed_len_x, fixed_len_y, indices


class LibSVMDataset(Dataset):
    """
    LibSVM Learning to Rank dataset.
    """
    def __init__(self, X, y, query_ids, transform=None):
        """
        :param X: scipy sparse matrix containing features of the dataset of shape [dataset_size, features_dim]
        :param y: ndarray containing target labels of shape [dataset_size]
        :param query_ids: ndarray containing group (slate) membership of dataset items of shape [dataset_size, features_dim]
        :param transform: a callable defining an optional transformation called on the dataset
        """
        X = X.toarray()

        groups = np.cumsum(np.unique(query_ids, return_counts=True)[1])

        self.X_by_qid = np.split(X, groups)[:-1]
        self.y_by_qid = np.split(y, groups)[:-1]

        self.longest_query_length = max([len(a) for a in self.X_by_qid])

        logger.info("loaded dataset with {} queries".format(len(self.X_by_qid)))
        logger.info("longest query had {} documents".format(self.longest_query_length))

        self.transform = transform

    @classmethod
    def from_svm_file(cls, svm_file_path, transform=None):
        """
        Instantiate a LibSVMDataset from a LibSVM file path.
        :param svm_file_path: LibSVM file path
        :param transform: a callable defining an optional transformation called on the dataset
        :return: LibSVMDataset instantiated from a given file and with an optional transformation defined
        """
        x, y, query_ids = load_svmlight_file(svm_file_path, query_id=True)
        logger.info("loaded dataset from {} and got x shape {}, y shape {} and query_ids shape {}".format(
            svm_file_path, x.shape, y.shape, query_ids.shape))
        return cls(x, y, query_ids, transform)

    def __len__(self):
        """
        :return: number of groups (slates) in the dataset
        """
        return len(self.X_by_qid)

    def __getitem__(self, idx):
        """
        :param idx: index of a group
        :return: ndarrays tuple containing features and labels of shapes [slate_length, features_dim] and [slate_length], respectively
        """
        X = self.X_by_qid[idx]
        y = self.y_by_qid[idx]

        sample = X, y

        if self.transform:
            sample = self.transform(sample)

        return sample

    @property
    def shape(self):
        """
        :return: shape of the dataset [batch_dim, document_dim, features_dim] where batch_dim is the number of groups
            (slates) and document_dim is the length of the longest group
        """
        batch_dim = len(self)
        document_dim = self.longest_query_length
        features_dim = self[0][0].shape[-1]
        return [batch_dim, document_dim, features_dim]


def load_libsvm_role(input_path: str, role: str) -> LibSVMDataset:
    """
    Helper function loading a LibSVMDataset of a specific role.

    The file can be located either in the local filesystem or in GCS.
    :param input_path: LibSVM file directory
    :param role: dataset role (file name without an extension)
    :return: LibSVMDataset from file {input_path}/{role}.txt
    """
    path = os.path.join(input_path, "{}.txt".format(role))
    logger.info("will load {} data from {}".format(role, path))
    with open_local_or_gs(path, "rb") as input_stream:
        ds = LibSVMDataset.from_svm_file(input_stream)
    logger.info("{} DS shape: {}".format(role, ds.shape))
    return ds


def fix_length_to_longest_slate(ds: LibSVMDataset) -> Compose:
    """
    Helper function returning a transforms.Compose object performing length fixing and tensor conversion.

    Length fixing operation will fix every slate's length to maximum length present in the LibSVMDataset.
    :param ds: LibSVMDataset to transform
    :return: transforms.Compose object
    """
    logger.info("Will pad to the longest slate: {}".format(ds.longest_query_length))
    return transforms.Compose([FixLength(int(ds.longest_query_length)), ToTensor()])


def load_libsvm_dataset(input_path: str, slate_length: int, validation_ds_role: str) \
        -> Tuple[LibSVMDataset, LibSVMDataset]:
    """
    Helper function loading a train LibSVMDataset and a specified validation LibSVMDataset.
    :param input_path: directory containing the LibSVM files
    :param slate_length: target slate length of the training dataset
    :param validation_ds_role: dataset role used for valdation (file name without an extension)
    :return: tuple of LibSVMDatasets containing train and validation datasets,
        where train slates are padded to slate_length and validation slates to val_ds.longest_query_length
    """
    train_ds = load_libsvm_dataset_role("train", input_path, slate_length)

    val_ds = load_libsvm_dataset_role(validation_ds_role, input_path, slate_length)

    return train_ds, val_ds


def load_libsvm_dataset_role(role: str, input_path: str, slate_length: int) -> LibSVMDataset:
    """
    Helper function loading a single role LibSVMDataset
    :param role: the role of the dataset - specifies file name and padding behaviour
    :param input_path: directory containing the LibSVM files
    :param slate_length: target slate length of the training dataset
    :return: loaded LibSVMDataset
    """
    ds = load_libsvm_role(input_path, role)
    if role == "train":
        ds.transform = transforms.Compose([FixLength(slate_length), ToTensor()])
    else:
        ds.transform = fix_length_to_longest_slate(ds)
    return ds


def create_data_loaders(train_ds: LibSVMDataset, val_ds: LibSVMDataset, num_workers: int, batch_size: int):
    """
    Helper function creating train and validation data loaders with specified number of workers and batch sizes.
    :param train_ds: LibSVMDataset train dataset
    :param val_ds: LibSVMDataset validation dataset
    :param num_workers: number of data loader workers
    :param batch_size: size of the batches returned by the data loaders
    :return: tuple containing train and validation DataLoader objects
    """
    # We are multiplying the batch size by the processing units count
    gpu_count = torch.cuda.device_count()
    total_batch_size = max(1, gpu_count) * batch_size
    logger.info("total batch size is {}".format(total_batch_size))

    # Please note that the batch size for validation dataloader is twice the total_batch_size
    train_dl = DataLoader(train_ds, batch_size=total_batch_size, num_workers=num_workers, shuffle=True)
    val_dl = DataLoader(val_ds, batch_size=total_batch_size, num_workers=num_workers, shuffle=False)
    return train_dl, val_dl



from urllib.parse import urlparse

import allrank.models.losses as losses
import numpy as np
import os
import torch
from allrank.config import Config
from allrank.data.dataset_loading import load_libsvm_dataset, create_data_loaders
from allrank.models.model import make_model
from allrank.models.model_utils import get_torch_device, CustomDataParallel
from allrank.training.train_utils import fit
from allrank.utils.command_executor import execute_command
from allrank.utils.experiments import dump_experiment_result, assert_expected_metrics
from allrank.utils.file_utils import create_output_dirs, PathsContainer, copy_local_to_gs
from allrank.utils.ltr_logging import init_logger
from allrank.utils.python_utils import dummy_context_mgr
from argparse import ArgumentParser, Namespace
from attr import asdict
from functools import partial
from pprint import pformat
from torch import optim


def parse_args() -> Namespace:
    parser = ArgumentParser("allRank")
    parser.add_argument("--job-dir", help="Base output path for all experiments", required=True)
    parser.add_argument("--run-id", help="Name of this run to be recorded (must be unique within output dir)",
                        required=True)
    parser.add_argument("--config-file-name", required=True, type=str, help="Name of json file with config")

    return parser.parse_args()


def run():
    # reproducibility
    torch.manual_seed(42)
    torch.cuda.manual_seed_all(42)
    np.random.seed(42)

    args = parse_args()

    paths = PathsContainer.from_args(args.job_dir, args.run_id, args.config_file_name)

    create_output_dirs(paths.output_dir)

    logger = init_logger(paths.output_dir)
    logger.info(f"created paths container {paths}")

    # read config
    config = Config.from_json(paths.config_path)
    logger.info("Config:\n {}".format(pformat(vars(config), width=1)))

    output_config_path = os.path.join(paths.output_dir, "used_config.json")
    execute_command("cp {} {}".format(paths.config_path, output_config_path))

    # train_ds, val_ds
    train_ds, val_ds = load_libsvm_dataset(
        input_path=config.data.path,
        slate_length=config.data.slate_length,
        validation_ds_role=config.data.validation_ds_role,
    )

    n_features = train_ds.shape[-1]
    assert n_features == val_ds.shape[-1], "Last dimensions of train_ds and val_ds do not match!"

    # train_dl, val_dl
    train_dl, val_dl = create_data_loaders(
        train_ds, val_ds, num_workers=config.data.num_workers, batch_size=config.data.batch_size)

    # gpu support
    dev = get_torch_device()
    logger.info("Model training will execute on {}".format(dev.type))

    # instantiate model
    model = make_model(n_features=n_features, **asdict(config.model, recurse=False))
    if torch.cuda.device_count() > 1:
        model = CustomDataParallel(model)
        logger.info("Model training will be distributed to {} GPUs.".format(torch.cuda.device_count()))
    model.to(dev)

    # load optimizer, loss and LR scheduler
    optimizer = getattr(optim, config.optimizer.name)(params=model.parameters(), **config.optimizer.args)
    loss_func = partial(getattr(losses, config.loss.name), **config.loss.args)
    if config.lr_scheduler.name:
        scheduler = getattr(optim.lr_scheduler, config.lr_scheduler.name)(optimizer, **config.lr_scheduler.args)
    else:
        scheduler = None

    with torch.autograd.detect_anomaly() if config.detect_anomaly else dummy_context_mgr():  # type: ignore
        # run training
        result = fit(
            model=model,
            loss_func=loss_func,
            optimizer=optimizer,
            scheduler=scheduler,
            train_dl=train_dl,
            valid_dl=val_dl,
            config=config,
            device=dev,
            output_dir=paths.output_dir,
            tensorboard_output_path=paths.tensorboard_output_path,
            **asdict(config.training)
        )

    dump_experiment_result(args, config, paths.output_dir, result)

    if urlparse(args.job_dir).scheme == "gs":
        copy_local_to_gs(paths.local_base_output_path, args.job_dir)

    assert_expected_metrics(result, config.expected_metrics)


if __name__ == "__main__":
    run()



from typing import Tuple, Union

import numpy as np
import torch
from scipy.spatial.distance import cdist

from allrank.click_models.base import ClickModel


class EverythingButDuplicatesClickModel(ClickModel):
    """
    This ClickModel clicks on every document, which was not previously clicked,
    if the distance between this document and any previous is larger than given margin in given metric
    """

    def __init__(self, duplicate_margin: float = 0, metric: str = "euclidean"):
        """

        :param duplicate_margin: a margin to tell whether a pair of documents is treated as a duplicate.
            If the distance is less than or equal to this value - this marks a duplicate
        :param metric: a metric in which pairwise distances are calculated
            (metric must be supported by `scipy.spatial.distance.cdist`)
        """
        self.duplicate_margin = duplicate_margin
        self.metric = metric

    def click(self, documents: Tuple[torch.Tensor, Union[torch.Tensor, np.ndarray]]) -> np.ndarray:
        X, y = documents
        dist = cdist(X, X, metric=self.metric)
        dist = np.triu(dist, k=1)
        np.fill_diagonal(dist, np.inf)
        indices = np.tril_indices(dist.shape[0])
        dist[indices] = np.inf
        return 1 * (dist > self.duplicate_margin).min(0)



import importlib
from typing import Any

from allrank.config import NameArgsConfig


def instantiate_from_recursive_name_args(name_args: NameArgsConfig):
    def instantiate_if_name_args(o: Any):
        if isinstance(o, NameArgsConfig):
            return instantiate_from_recursive_name_args(o)
        elif isinstance(o, dict) and {"name", "args"} == o.keys():
            return instantiate_from_recursive_name_args(NameArgsConfig(**o))
        else:
            return o

    instantiated_args = dict([(k, instantiate_if_name_args(v)) for k, v in name_args.args.items()])
    return instantiate_class(name_args.name, **instantiated_args)


def instantiate_class(full_name: str, **kwargs):
    module_name, class_name = full_name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_(**kwargs)



import torch.nn as nn
from attr import asdict

from allrank.models.transformer import make_transformer
from allrank.utils.python_utils import instantiate_class


def first_arg_id(x, *y):
    return x


class FCModel(nn.Module):
    """
    This class represents a fully connected neural network model with given layer sizes and activation function.
    """
    def __init__(self, sizes, input_norm, activation, dropout, n_features):
        """
        :param sizes: list of layer sizes (excluding the input layer size which is given by n_features parameter)
        :param input_norm: flag indicating whether to perform layer normalization on the input
        :param activation: name of the PyTorch activation function, e.g. Sigmoid or Tanh
        :param dropout: dropout probability
        :param n_features: number of input features
        """
        super(FCModel, self).__init__()
        sizes.insert(0, n_features)
        layers = [nn.Linear(size_in, size_out) for size_in, size_out in zip(sizes[:-1], sizes[1:])]
        self.input_norm = nn.LayerNorm(n_features) if input_norm else nn.Identity()
        self.activation = nn.Identity() if activation is None else instantiate_class(
            "torch.nn.modules.activation", activation)
        self.dropout = nn.Dropout(dropout or 0.0)
        self.output_size = sizes[-1]

        self.layers = nn.ModuleList(layers)

    def forward(self, x):
        """
        Forward pass through the FCModel.
        :param x: input of shape [batch_size, slate_length, self.layers[0].in_features]
        :return: output of shape [batch_size, slate_length, self.output_size]
        """
        x = self.input_norm(x)
        for layer in self.layers:
            x = self.dropout(self.activation(layer(x)))
        return x


class LTRModel(nn.Module):
    """
    This class represents a full neural Learning to Rank model with a given encoder model.
    """
    def __init__(self, input_layer, encoder, output_layer):
        """
        :param input_layer: the input block (e.g. FCModel)
        :param encoder: the encoding block (e.g. transformer.Encoder)
        :param output_layer: the output block (e.g. OutputLayer)
        """
        super(LTRModel, self).__init__()
        self.input_layer = input_layer if input_layer else nn.Identity()
        self.encoder = encoder if encoder else first_arg_id
        self.output_layer = output_layer

    def prepare_for_output(self, x, mask, indices):
        """
        Forward pass through the input layer and encoder.
        :param x: input of shape [batch_size, slate_length, input_dim]
        :param mask: padding mask of shape [batch_size, slate_length]
        :param indices: original item ranks used in positional encoding, shape [batch_size, slate_length]
        :return: encoder output of shape [batch_size, slate_length, encoder_output_dim]
        """
        return self.encoder(self.input_layer(x), mask, indices)

    def forward(self, x, mask, indices):
        """
        Forward pass through the whole LTRModel.
        :param x: input of shape [batch_size, slate_length, input_dim]
        :param mask: padding mask of shape [batch_size, slate_length]
        :param indices: original item ranks used in positional encoding, shape [batch_size, slate_length]
        :return: model output of shape [batch_size, slate_length, output_dim]
        """
        return self.output_layer(self.prepare_for_output(x, mask, indices))

    def score(self, x, mask, indices):
        """
        Forward pass through the whole LTRModel and item scoring.

        Used when evaluating listwise metrics in the training loop.
        :param x: input of shape [batch_size, slate_length, input_dim]
        :param mask: padding mask of shape [batch_size, slate_length]
        :param indices: original item ranks used in positional encoding, shape [batch_size, slate_length]
        :return: scores of shape [batch_size, slate_length]
        """
        return self.output_layer.score(self.prepare_for_output(x, mask, indices))


class OutputLayer(nn.Module):
    """
    This class represents an output block reducing the output dimensionality to d_output.
    """
    def __init__(self, d_model, d_output, output_activation=None):
        """
        :param d_model: dimensionality of the output layer input
        :param d_output: dimensionality of the output layer output
        :param output_activation: name of the PyTorch activation function used before scoring, e.g. Sigmoid or Tanh
        """
        super(OutputLayer, self).__init__()
        self.activation = nn.Identity() if output_activation is None else instantiate_class(
            "torch.nn.modules.activation", output_activation)
        self.d_output = d_output
        self.w_1 = nn.Linear(d_model, d_output)

    def forward(self, x):
        """
        Forward pass through the OutputLayer.
        :param x: input of shape [batch_size, slate_length, self.d_model]
        :return: output of shape [batch_size, slate_length, self.d_output]
        """
        return self.activation(self.w_1(x).squeeze(dim=2))

    def score(self, x):
        """
        Forward pass through the OutputLayer and item scoring by summing the individual outputs if d_output > 1.
        :param x: input of shape [batch_size, slate_length, self.d_model]
        :return: output of shape [batch_size, slate_length]
        """
        if self.d_output > 1:
            return self.forward(x).sum(-1)
        else:
            return self.forward(x)


def make_model(fc_model, transformer, post_model, n_features):
    """
    Helper function for instantiating LTRModel.
    :param fc_model: FCModel used as input block
    :param transformer: transformer Encoder used as encoder block
    :param post_model: parameters dict for OutputModel output block (excluding d_model)
    :param n_features: number of input features
    :return: LTR model instance
    """
    if fc_model:
        fc_model = FCModel(**fc_model, n_features=n_features)  # type: ignore
    d_model = n_features if not fc_model else fc_model.output_size
    if transformer:
        transformer = make_transformer(n_features=d_model, **asdict(transformer, recurse=False))  # type: ignore
    model = LTRModel(fc_model, transformer, OutputLayer(d_model, **post_model))

    # Initialize parameters with Glorot / fan_avg.
    for p in model.parameters():
        if p.dim() > 1:
            nn.init.xavier_uniform_(p)
    return model



from typing import Tuple, Dict, List, Generator

import torch
from torch.utils.data.dataloader import DataLoader

import allrank.models.losses as losses
from allrank.config import Config
from allrank.data.dataset_loading import LibSVMDataset
from allrank.models.metrics import ndcg, dcg
from allrank.models.model import LTRModel
from allrank.models.model_utils import get_torch_device


def rank_slates(datasets: Dict[str, LibSVMDataset], model: LTRModel, config: Config) \
        -> Dict[str, Tuple[torch.Tensor, torch.Tensor]]:
    """
    Ranks given datasets according to a given model

    :param datasets: dictionary of role -> dataset that will be ranked
    :param model: a model to use for scoring documents
    :param config: config for DataLoaders
    :return: dictionary of role -> ranked dataset
        every dataset is a Tuple of torch.Tensor - storing X and y in the descending order of the scores.
    """

    dataloaders = {role: __create_data_loader(ds, config) for role, ds in datasets.items()}

    ranked_slates = {role: __rank_slates(dl, model) for role, dl in dataloaders.items()}

    return ranked_slates


def __create_data_loader(ds: LibSVMDataset, config: Config) -> DataLoader:
    return DataLoader(ds, batch_size=config.data.batch_size, num_workers=config.data.num_workers, shuffle=False)


def __rank_slates(dataloader: DataLoader, model: LTRModel) -> Tuple[torch.Tensor, torch.Tensor]:
    reranked_X = []
    reranked_y = []
    model.eval()
    device = get_torch_device()
    with torch.no_grad():
        for xb, yb, _ in dataloader:
            X = xb.type(torch.float32).to(device=device)
            y_true = yb.to(device=device)

            input_indices = torch.ones_like(y_true).type(torch.long)
            mask = (y_true == losses.PADDED_Y_VALUE)
            scores = model.score(X, mask, input_indices)

            scores[mask] = float('-inf')

            _, indices = scores.sort(descending=True, dim=-1)
            indices_X = torch.unsqueeze(indices, -1).repeat_interleave(X.shape[-1], -1)
            reranked_X.append(torch.gather(X, dim=1, index=indices_X).cpu())
            reranked_y.append(torch.gather(y_true, dim=1, index=indices).cpu())

    combined_X = torch.cat(reranked_X)
    combined_y = torch.cat(reranked_y)
    return combined_X, combined_y


def __clicked_ndcg(ordered_clicks: List[int]) -> float:
    return ndcg(torch.arange(start=len(ordered_clicks), end=0, step=-1, dtype=torch.float32)[None, :],
                torch.tensor(ordered_clicks)[None, :]).item()


def __clicked_dcg(ordered_clicks: List[int]) -> float:
    return dcg(torch.arange(start=len(ordered_clicks), end=0, step=-1, dtype=torch.float32)[None, :],
               torch.tensor(ordered_clicks)[None, :]).item()


def metrics_on_clicked_slates(clicked_slates: Tuple[List[torch.Tensor], List[List[int]]]) \
        -> Generator[Dict[str, float], None, None]:
    Xs, ys = clicked_slates
    for X, y in zip(Xs, ys):
        yield {
            "slate_length": len(y),
            "no_of_clicks": sum(y > 0),  # type: ignore
            "dcg": __clicked_dcg(y),
            "ndcg": __clicked_ndcg(y)
        }



from itertools import product

import torch
from torch.nn import BCEWithLogitsLoss

from allrank.data.dataset_loading import PADDED_Y_VALUE


def rankNet_weightByGTDiff(y_pred, y_true, padded_value_indicator=PADDED_Y_VALUE):
    """
    Wrapper for RankNet employing weighing by the differences of ground truth values.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: loss value, a torch.Tensor
    """
    return rankNet(y_pred, y_true, padded_value_indicator, weight_by_diff=True)


def rankNet_weightByGTDiff_pow(y_pred, y_true, padded_value_indicator=PADDED_Y_VALUE):
    """
    Wrapper for RankNet employing weighing by the squared differences of ground truth values.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: loss value, a torch.Tensor
    """
    return rankNet(y_pred, y_true, padded_value_indicator, weight_by_diff=False, weight_by_diff_powed=True)


def rankNet(y_pred, y_true, padded_value_indicator=PADDED_Y_VALUE, weight_by_diff=False, weight_by_diff_powed=False):
    """
    RankNet loss introduced in "Learning to Rank using Gradient Descent".
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param weight_by_diff: flag indicating whether to weight the score differences by ground truth differences.
    :param weight_by_diff_powed: flag indicating whether to weight the score differences by the squared ground truth differences.
    :return: loss value, a torch.Tensor
    """
    y_pred = y_pred.clone()
    y_true = y_true.clone()

    mask = y_true == padded_value_indicator
    y_pred[mask] = float('-inf')
    y_true[mask] = float('-inf')

    # here we generate every pair of indices from the range of document length in the batch
    document_pairs_candidates = list(product(range(y_true.shape[1]), repeat=2))

    pairs_true = y_true[:, document_pairs_candidates]
    selected_pred = y_pred[:, document_pairs_candidates]

    # here we calculate the relative true relevance of every candidate pair
    true_diffs = pairs_true[:, :, 0] - pairs_true[:, :, 1]
    pred_diffs = selected_pred[:, :, 0] - selected_pred[:, :, 1]

    # here we filter just the pairs that are 'positive' and did not involve a padded instance
    # we can do that since in the candidate pairs we had symetric pairs so we can stick with
    # positive ones for a simpler loss function formulation
    the_mask = (true_diffs > 0) & (~torch.isinf(true_diffs))

    pred_diffs = pred_diffs[the_mask]

    weight = None
    if weight_by_diff:
        abs_diff = torch.abs(true_diffs)
        weight = abs_diff[the_mask]
    elif weight_by_diff_powed:
        true_pow_diffs = torch.pow(pairs_true[:, :, 0], 2) - torch.pow(pairs_true[:, :, 1], 2)
        abs_diff = torch.abs(true_pow_diffs)
        weight = abs_diff[the_mask]

    # here we 'binarize' true relevancy diffs since for a pairwise loss we just need to know
    # whether one document is better than the other and not about the actual difference in
    # their relevancy levels
    true_diffs = (true_diffs > 0).type(torch.float32)
    true_diffs = true_diffs[the_mask]

    return BCEWithLogitsLoss(weight=weight)(pred_diffs, true_diffs)



import numpy as np

from allrank.click_models.base import FixedClickModel, RandomClickModel
from allrank.click_models.click_utils import MaskedRemainMasked
from allrank.data.dataset_loading import PADDED_Y_VALUE
from tests.click_models import click


def test_masked_should_remain_masked():
    click_model = MaskedRemainMasked(FixedClickModel(click_positions=[1]))
    assert click(click_model, np.ones((3, 1)), np.array([0, 0, PADDED_Y_VALUE])) == [0, 1, PADDED_Y_VALUE]


def test_inner_click_model_should_just_get_unmasked_docs():
    np.random.seed(42)
    click_model = MaskedRemainMasked(RandomClickModel(n_clicks=1))
    clicks = click(click_model, np.ones((5, 1)), np.array([0, PADDED_Y_VALUE, PADDED_Y_VALUE, PADDED_Y_VALUE, PADDED_Y_VALUE]))
    assert clicks == [1, PADDED_Y_VALUE, PADDED_Y_VALUE, PADDED_Y_VALUE, PADDED_Y_VALUE]



import numpy as np

from allrank.click_models.duplicate_aware import EverythingButDuplicatesClickModel
from tests.click_models import click


def test_feature_click_model_everything():
    click_model = EverythingButDuplicatesClickModel()
    assert click(click_model, np.array([[0, 1]]), []) == [1]
    assert click(click_model, np.array([[1, 1], [1, 0]]), []) == [1, 1]
    assert click(click_model, np.array([[1, 1], [1, 0], [0, 0]]), []) == [1, 1, 1]


def test_feature_click_model_except_exact_duplicates():
    click_model = EverythingButDuplicatesClickModel()
    assert click(click_model, np.array([[0, 1]]), []) == [1]
    assert click(click_model, np.array([[1, 1], [1, 1]]), []) == [1, 0]
    assert click(click_model, np.array([[1, 1], [1, 1], [0, 0]]), []) == [1, 0, 1]


def test_feature_click_model_except_near_duplicates():
    click_model = EverythingButDuplicatesClickModel(0.1)
    assert click(click_model, np.array([[0, 1]]), []) == [1]
    assert click(click_model, np.array([[1, 1], [1, 1]]), []) == [1, 0]
    assert click(click_model, np.array([[1, 1], [1, 0.99], [1, 0.8]]), []) == [1, 0, 1]



import torch

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import DEFAULT_EPS


def listMLE(y_pred, y_true, eps=DEFAULT_EPS, padded_value_indicator=PADDED_Y_VALUE):
    """
    ListMLE loss introduced in "Listwise Approach to Learning to Rank - Theory and Algorithm".
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param eps: epsilon value, used for numerical stability
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: loss value, a torch.Tensor
    """
    # shuffle for randomised tie resolution
    random_indices = torch.randperm(y_pred.shape[-1])
    y_pred_shuffled = y_pred[:, random_indices]
    y_true_shuffled = y_true[:, random_indices]

    y_true_sorted, indices = y_true_shuffled.sort(descending=True, dim=-1)

    mask = y_true_sorted == padded_value_indicator

    preds_sorted_by_true = torch.gather(y_pred_shuffled, dim=1, index=indices)
    preds_sorted_by_true[mask] = float("-inf")

    max_pred_values, _ = preds_sorted_by_true.max(dim=1, keepdim=True)

    preds_sorted_by_true_minus_max = preds_sorted_by_true - max_pred_values

    cumsums = torch.cumsum(preds_sorted_by_true_minus_max.exp().flip(dims=[1]), dim=1).flip(dims=[1])

    observation_loss = torch.log(cumsums + eps) - preds_sorted_by_true_minus_max

    observation_loss[mask] = 0.0

    return torch.mean(torch.sum(observation_loss, dim=1))



import torch

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.metrics import mrr


def mrr_wrap_single_slate(y_pred, y_true, ats=[10]):
    return mrr_wrap_multiple_slates([y_pred], [y_true], ats)


def mrr_wrap_multiple_slates(y_pred, y_true, ats=[10]):
    return mrr(torch.tensor(y_pred), torch.tensor(y_true), ats=ats).numpy()


def single_slate_and_ats(result):
    return result[0][0]


def test_mrr_simple_1():
    y_pred = [0.5, 0.2]
    y_true = [1.0, 0.0]

    result = mrr_wrap_single_slate(y_pred, y_true)

    assert single_slate_and_ats(result) == 1.0


def test_mrr_simple_no_ats():
    y_pred = [0.5, 0.2]
    y_true = [1.0, 0.0]

    result = mrr_wrap_single_slate(y_pred, y_true, ats=None)

    assert single_slate_and_ats(result) == 1.0


def test_mrr_simple_2():
    y_pred = [0.5, 0.2]
    y_true = [0.0, 1.0]

    result = mrr_wrap_single_slate(y_pred, y_true)

    assert single_slate_and_ats(result) == 0.5


def test_mrr_multiple_slates():
    y_pred_1 = [0.2, 0.5]
    y_pred_05 = [0.5, 0.2]
    y_true = [0.0, 1.0]

    result = mrr_wrap_multiple_slates([y_pred_1, y_pred_05], [y_true, y_true])

    assert result[0][0] == 1.0
    assert result[1][0] == 0.5


def test_mrr_multiple_ats():
    y_pred = [0.5, 0.2]
    y_true = [0.0, 1.0]

    result = mrr_wrap_single_slate(y_pred, y_true, ats=[1, 2])

    assert result[0][0] == 0.0
    assert result[0][1] == 0.5


def test_mrr_multiple_slates_multiple_ats():
    y_pred_1 = [0.2, 0.5]
    y_pred_05 = [0.5, 0.2]
    y_true = [0.0, 1.0]

    result = mrr_wrap_multiple_slates([y_pred_1, y_pred_05], [y_true, y_true], ats=[1, 2])

    assert result[0][0] == 1.0
    assert result[0][1] == 1.0
    assert result[1][0] == 0.0
    assert result[1][1] == 0.5


def test_mrr_zero_when_no_relevant():
    y_pred = [0.5, 0.2]
    y_true = [0.0, 0.0]

    result = mrr_wrap_single_slate(y_pred, y_true)

    assert single_slate_and_ats(result) == 0.0


def test_mrr_with_padded_input():
    y_pred = [0.5, 0.2, 1.0]
    y_true = [1.0, 0.0, PADDED_Y_VALUE]

    result = mrr_wrap_single_slate(y_pred, y_true)

    assert single_slate_and_ats(result) == 1.0


def test_mrr_with_padded_input_2():
    y_pred = [0.5, 0.2, 1.0]
    y_true = [0.0, 1.0, PADDED_Y_VALUE]

    result = mrr_wrap_single_slate(y_pred, y_true)

    assert single_slate_and_ats(result) == 0.5



import math

import torch
from pytest import approx

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import listMLE


def listmle_wrap(y_pred, y_true):
    return listMLE(torch.tensor([y_pred]), torch.tensor([y_true])).item()


def test_listmle_ignores_padded_value():
    y_pred = [0.5, 0.3, 0.5]
    y_true = [1.0, 0.0, PADDED_Y_VALUE]

    result = listmle_wrap(y_pred, y_true)
    expected = 0.5981389284133911

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))



import logging
import os
import sys


def init_logger(output_dir: str) -> logging.Logger:
    log_format = "[%(levelname)s] %(asctime)s - %(message)s"
    log_dateformat = "%Y-%m-%d %H:%M:%S"

    logging.basicConfig(format=log_format, datefmt=log_dateformat, stream=sys.stdout, level=logging.INFO)

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    # create a file handler
    handler = logging.FileHandler(os.path.join(output_dir, "training.log"))
    handler.setLevel(logging.INFO)

    # create a logging format
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def get_logger() -> logging.Logger:
    return logging.getLogger(__name__)



import math

import numpy as np
import torch
from pytest import approx

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import pointwise_rmse


def loss_wrap(y_pred, y_true, no_of_levels):
    return pointwise_rmse(torch.tensor([y_pred]), torch.tensor([y_true]), no_of_levels).item()


def test_pointwise_simple():
    y_pred = [0.5, 0.2]
    y_true = [1.0, 0.0]

    result = loss_wrap(y_pred, y_true, 1)
    expected = math.sqrt(np.mean([0.5 ** 2, 0.2 ** 2]))

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_pointwise_simple_padded():
    y_pred = [0.5, 0.2, 0.5]
    y_true = [1.0, 0.0, PADDED_Y_VALUE]

    result = loss_wrap(y_pred, y_true, 1)
    expected = math.sqrt(np.mean([0.5 ** 2, 0.2 ** 2]))

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))


def test_pointwise_multiple_levels():
    y_pred = [0.5, 0.2, 0.7, 0.8]
    y_true = [1.0, 0.0, 2.0, 3.0]

    result = loss_wrap(y_pred, y_true, 3)
    expected = math.sqrt(np.mean([0.5 ** 2, 0.6 ** 2, 0.1 ** 2, 0.6 ** 2]))

    assert not math.isnan(result) and not math.isinf(result)
    assert (result == approx(expected))



from typing import List, Tuple, Union

import numpy as np
import torch

from allrank.click_models.base import ClickModel
from allrank.data.dataset_loading import PADDED_Y_VALUE


def click_on_slates(slates: Union[Tuple[np.ndarray, np.ndarray], Tuple[torch.Tensor, torch.Tensor]],
                    click_model: ClickModel, include_empty: bool) -> Tuple[List[Union[np.ndarray, torch.Tensor]], List[List[int]]]:
    """
    This metod runs a click model on a list of slates and returns new slates with `y` taken from clicks

    :param slates: a Tuple of X, y:
        X being a list of slates represented by document vectors
        y being a list of slates represented by document relevancies
    :param click_model: a click model to be applied to every slate
    :param include_empty: if True - will return even slates that didn't get any click
    :return: Tuple of X, clicks, X representing the same document vectors as input 'X', clicks representing click mask for every slate
    """
    X, y = slates
    clicks = [MaskedRemainMasked(click_model).click(slate) for slate in zip(X, y)]
    X_with_clicks = [[X, slate_clicks] for X, slate_clicks in list(zip(X, clicks)) if
                     (np.sum(slate_clicks > 0) > 0 or include_empty)]
    return_X, clicks = map(list, zip(*X_with_clicks))  # type: ignore
    return return_X, clicks  # type: ignore


class MaskedRemainMasked(ClickModel):
    """
    This click model wraps another click model and:
      1. ensures inner click model do not get documents that were padded
      2. ensures padded documents get '-1' in 'clicked' vector
    """

    def __init__(self, inner_click_model: ClickModel):
        """

        :param inner_click_model: a click model that is run on the list of non-padded documents
        """
        self.inner_click_model = inner_click_model

    def click(self, documents: Union[Tuple[np.ndarray, np.ndarray], Tuple[torch.Tensor, torch.Tensor]]) -> np.ndarray:
        X, y = documents
        padded_values_mask = y == PADDED_Y_VALUE
        real_X = X[~padded_values_mask]
        real_y = y[~padded_values_mask]
        clicks = self.inner_click_model.click((real_X, real_y))
        final_clicks = np.zeros_like(y)
        final_clicks[padded_values_mask] = PADDED_Y_VALUE
        final_clicks[~padded_values_mask] = clicks
        return final_clicks



import os
from argparse import ArgumentParser, Namespace
from pprint import pformat
from urllib.parse import urlparse

import numpy as np
import pandas as pd
import torch
from attr import asdict

from allrank.click_models.click_utils import click_on_slates
from allrank.config import Config
from allrank.data.dataset_loading import load_libsvm_dataset_role
from allrank.data.dataset_saving import write_to_libsvm_without_masked
from allrank.inference.inference_utils import rank_slates, metrics_on_clicked_slates
from allrank.models.model import make_model
from allrank.models.model_utils import get_torch_device, CustomDataParallel, load_state_dict_from_file
from allrank.utils.args_utils import split_as_strings
from allrank.utils.command_executor import execute_command
from allrank.utils.config_utils import instantiate_from_recursive_name_args
from allrank.utils.file_utils import create_output_dirs, PathsContainer, copy_local_to_gs
from allrank.utils.ltr_logging import init_logger
from allrank.utils.python_utils import all_equal


def parse_args() -> Namespace:
    parser = ArgumentParser("allRank rank and apply click model")
    parser.add_argument("--job-dir", help="Base output path for all experiments", required=True)
    parser.add_argument("--run-id", help="Name of this run to be recorded (must be unique within output dir)",
                        required=True)
    parser.add_argument("--config-file-name", required=True, type=str, help="Name of json file with model config")
    parser.add_argument("--input-model-path", required=True, type=str, help="Path to the model to read weights")
    parser.add_argument("--roles", required=True, type=split_as_strings,
                        help="List of comma-separated dataset roles to load and process")

    return parser.parse_args()


def run():
    # reproducibility
    torch.manual_seed(42)
    torch.cuda.manual_seed_all(42)
    np.random.seed(42)

    args = parse_args()

    paths = PathsContainer.from_args(args.job_dir, args.run_id, args.config_file_name)

    os.makedirs(paths.base_output_path, exist_ok=True)

    create_output_dirs(paths.output_dir)
    logger = init_logger(paths.output_dir)

    logger.info("will save data in {output_dir}".format(output_dir=paths.base_output_path))

    # read config
    config = Config.from_json(paths.config_path)
    logger.info("Config:\n {}".format(pformat(vars(config), width=1)))

    output_config_path = os.path.join(paths.output_dir, "used_config.json")
    execute_command("cp {} {}".format(paths.config_path, output_config_path))

    datasets = {role: load_libsvm_dataset_role(role, config.data.path, config.data.slate_length) for role in args.roles}

    n_features = [ds.shape[-1] for ds in datasets.values()]
    assert all_equal(n_features), f"Last dimensions of datasets must match but got {n_features}"

    # gpu support
    dev = get_torch_device()
    logger.info("Will use device {}".format(dev.type))

    # instantiate model
    model = make_model(n_features=n_features[0], **asdict(config.model, recurse=False))

    model.load_state_dict(load_state_dict_from_file(args.input_model_path, dev))
    logger.info(f"loaded model weights from {args.input_model_path}")

    if torch.cuda.device_count() > 1:
        model = CustomDataParallel(model)
        logger.info("Model training will be distributed to {} GPUs.".format(torch.cuda.device_count()))
    model.to(dev)

    assert config.click_model is not None, "click_model must be defined in config for this run"
    click_model = instantiate_from_recursive_name_args(name_args=config.click_model)

    ranked_slates = rank_slates(datasets, model, config)

    clicked_slates = {role: click_on_slates(slates, click_model, include_empty=False) for role, slates in ranked_slates.items()}

    # save clickthrough datasets
    for role, slates in clicked_slates.items():
        write_to_libsvm_without_masked(os.path.join(paths.output_dir, f"{role}.txt"), *slates)  # type: ignore

    # calculate metrics
    metered_slates = {role: metrics_on_clicked_slates(slates) for role, slates in clicked_slates.items()}  # type: ignore

    for role, metrics in metered_slates.items():
        metrics_df = pd.DataFrame(metrics)
        logger.info(f"{role} metrics summary:")
        logger.info(metrics_df.mean())
        metrics_df.to_csv(os.path.join(paths.output_dir, f"{role}_metrics.csv"), index=False)
        pd.DataFrame(metrics_df.mean()).T.to_csv(os.path.join(paths.output_dir, f"{role}_metrics_mean.csv"), index=False)

    if urlparse(args.job_dir).scheme == "gs":
        copy_local_to_gs(paths.local_base_output_path, args.job_dir)


if __name__ == "__main__":
    run()



# coding=utf-8

import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

reqs = [
    "torch>=1.6.0, <=1.8.0",
    "torchvision>=0.7.0, <= 0.9.0",
    "scikit-learn>=0.23.0",
    "pandas>=1.0.5",
    "numpy>=1.18.5",
    "scipy>=1.4.1",
    "attrs>=19.3.0",
    "flatten_dict>=0.3.0",
    "tensorboardX>=2.1.0",
    "gcsfs==0.6.2",
    "google-auth==1.27.1"
]

setup(
    name="allRank",
    version="1.4.2",
    description="allRank is a framework for training learning-to-rank neural models",
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache 2",
    url="https://github.com/allegro/allRank",
    install_requires=reqs,
    author_email="allrank@allegro.pl",
    packages=find_packages(exclude=["tests"]),
    package_data={"allrank": ["config.json"]},
    entry_points={"console_scripts": ['allRank = allrank.main:run']},
    zip_safe=False,
)



import numpy as np

from allrank.click_models.base import RandomClickModel, OnlyRelevantClickModel
from allrank.click_models.click_utils import click_on_slates


def test_click_on_slates():
    np.random.seed(42)

    n_slates = 5
    n_docs_per_slate = 5
    n_dimensions = 10

    X = np.random.rand(n_slates, n_docs_per_slate, n_dimensions).astype(np.float32)
    y = np.vstack([np.random.randint(0, 4, size=len(x)) for x in X])

    n_clicks = 2
    click_model = RandomClickModel(n_clicks)

    slates_X, slates_y = click_on_slates((X, y), click_model, True)

    assert len(slates_X) == X.shape[0]
    assert (slates_X == X).all()
    assert len(slates_y) == len(y)
    assert (np.sum(slates_y, axis=1) == np.repeat(n_clicks, n_slates)).all()


def test_click_on_slates_without_empty():
    np.random.seed(42)

    X = np.array([[[-1.0]], [[1.0]]])
    y = np.vstack([np.array([0]), np.array([1])])

    click_model = OnlyRelevantClickModel(1)

    slates_X, slates_y = click_on_slates((X, y), click_model, include_empty=False)

    assert len(slates_X) == X[1:].shape[0]
    assert (slates_X == X[1:]).all()
    assert slates_y == [[1]]



import torch
from torch.nn import BCELoss

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.model_utils import get_torch_device


def with_ordinals(y, n, padded_value_indicator=PADDED_Y_VALUE):
    """
    Helper function for ordinal loss, transforming input labels to ordinal values.
    :param y: labels, shape [batch_size, slate_length]
    :param n: number of ordinals
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: ordinals, shape [batch_size, slate_length, n]
    """
    dev = get_torch_device()
    one_to_n = torch.arange(start=1, end=n + 1, dtype=torch.float, device=dev)
    unsqueezed = y.unsqueeze(2).repeat(1, 1, n)
    mask = unsqueezed == padded_value_indicator
    ordinals = (unsqueezed >= one_to_n).type(torch.float)
    ordinals[mask] = padded_value_indicator
    return ordinals


def ordinal(y_pred, y_true, n, padded_value_indicator=PADDED_Y_VALUE):
    """
    Ordinal loss.
    :param y_pred: predictions from the model, shape [batch_size, slate_length, n]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param n: number of ordinal values, int
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: loss value, a torch.Tensor
    """
    device = get_torch_device()

    y_pred = y_pred.clone()
    y_true = with_ordinals(y_true.clone(), n)

    mask = y_true == padded_value_indicator
    valid_mask = y_true != padded_value_indicator

    ls = BCELoss(reduction='none')(y_pred, y_true)
    ls[mask] = 0.0

    document_loss = torch.sum(ls, dim=2)
    sum_valid = torch.sum(valid_mask, dim=2).type(torch.float32) > torch.tensor(0.0, dtype=torch.float32, device=device)

    loss_output = torch.sum(document_loss) / torch.sum(sum_valid)

    return loss_output



from typing import Union, List

import numpy as np
import torch

from allrank.click_models.base import ClickModel


def click(click_model: ClickModel, X: Union[list, np.ndarray], y: Union[list, np.ndarray]) -> List[int]:
    clicks = click_model.click((torch.tensor(X), torch.tensor(y)))
    assert isinstance(clicks, np.ndarray)
    return clicks.tolist()



import json
from collections import defaultdict
from typing import Dict, List, Optional

from attr import attrib, attrs


@attrs
class TransformerConfig:
    N = attrib(type=int)
    d_ff = attrib(type=int)
    h = attrib(type=int)
    positional_encoding = attrib(type=dict)
    dropout = attrib(type=float)


@attrs
class FCConfig:
    sizes = attrib(type=List[int])
    input_norm = attrib(type=bool)
    activation = attrib(type=str)
    dropout = attrib(type=float)


@attrs
class PostModelConfig:
    d_output = attrib(type=int)
    output_activation = attrib(type=str)


@attrs
class ModelConfig:
    fc_model = attrib(type=FCConfig)
    transformer = attrib(type=TransformerConfig)
    post_model = attrib(type=PostModelConfig)


@attrs
class PositionalEncoding:
    strategy = attrib(type=str)
    max_indices = attrib(type=int)


@attrs
class DataConfig:
    path = attrib(type=str)
    num_workers = attrib(type=int)
    batch_size = attrib(type=int)
    slate_length = attrib(type=int)
    validation_ds_role = attrib(type=str)


@attrs
class TrainingConfig:
    epochs = attrib(type=int)
    gradient_clipping_norm = attrib(type=float)
    early_stopping_patience = attrib(type=int, default=0)


@attrs
class NameArgsConfig:
    name = attrib(type=str)
    args = attrib(type=dict)


@attrs
class Config:
    model = attrib(type=ModelConfig)
    data = attrib(type=DataConfig)
    optimizer = attrib(type=NameArgsConfig)
    training = attrib(type=TrainingConfig)
    loss = attrib(type=NameArgsConfig)
    metrics = attrib(type=Dict[str, List[int]])
    lr_scheduler = attrib(type=NameArgsConfig)
    val_metric = attrib(type=str, default=None)
    expected_metrics = attrib(type=Dict[str, Dict[str, float]], default={})
    detect_anomaly = attrib(type=bool, default=False)
    click_model = attrib(type=Optional[NameArgsConfig], default=None)

    @classmethod
    def from_json(cls, config_path):
        with open(config_path) as config_file:
            config = json.load(config_file)
            return Config.from_dict(config)

    @classmethod
    def from_dict(cls, config):
        config["model"] = ModelConfig(**config["model"])
        if config["model"].transformer:
            config["model"].transformer = TransformerConfig(**config["model"].transformer)
            if config["model"].transformer.positional_encoding:
                config["model"].transformer.positional_encoding = PositionalEncoding(
                    **config["model"].transformer.positional_encoding)
        config["data"] = DataConfig(**config["data"])
        config["optimizer"] = NameArgsConfig(**config["optimizer"])
        config["training"] = TrainingConfig(**config["training"])
        config["metrics"] = cls._parse_metrics(config["metrics"])
        config["lr_scheduler"] = NameArgsConfig(**config["lr_scheduler"])
        config["loss"] = NameArgsConfig(**config["loss"])
        if "click_model" in config.keys():
            config["click_model"] = NameArgsConfig(**config["click_model"])
        return cls(**config)

    @staticmethod
    def _parse_metrics(metrics):
        metrics_dict = defaultdict(list)  # type: Dict[str, list]
        for metric_string in metrics:
            try:
                name, at = metric_string.split("_")
                metrics_dict[name].append(int(at))
            except (ValueError, TypeError):
                raise MetricConfigError(
                    metric_string,
                    "Wrong formatting of metric in config. Expected format: <name>_<at> where name is valid metric name and at is and int")
        return metrics_dict


class MetricConfigError(Exception):
    pass



import numpy as np

from allrank.click_models.cascade_models import BaseCascadeModel
from tests.click_models import click


def test_base_cascade_model_no_eta():
    click_model = BaseCascadeModel(0.0, 1)
    assert click(click_model, [], [1]) == [1]
    assert click(click_model, [], [1, 2]) == [1, 1]
    assert click(click_model, [], [1, 2, 3]) == [1, 1, 1]


def test_base_cascade_model_below_threshold():
    y = [1, 2, 0, 4, 3]
    assert click(BaseCascadeModel(0.0, 1), [], y) == [1, 1, 0, 1, 1]
    assert click(BaseCascadeModel(0.0, 2), [], y) == [0, 1, 0, 1, 1]
    assert click(BaseCascadeModel(0.0, 4), [], y) == [0, 0, 0, 1, 0]


def test_base_cascade_model_eta():
    np.random.seed(42)
    click_model_1 = BaseCascadeModel(0.3, 1)
    click_model_2 = BaseCascadeModel(0.5, 1)
    assert click(click_model_1, [], [1, 2]) == [1, 0]
    assert click(click_model_1, [], [1, 2, 3]) == [1, 1, 1]
    assert click(click_model_1, [], [1, 2, 3, 4]) == [1, 1, 0, 1]
    assert click(click_model_2, [], [1, 2]) == [1, 1]
    assert click(click_model_2, [], [1, 2, 3]) == [1, 0, 1]
    assert click(click_model_2, [], [1, 2, 3, 4]) == [1, 1, 1, 0]


def test_base_cascade_model_obs_irrelevant():
    np.random.seed(42)
    y = [1, 2, 0, 4, 3]
    assert click(BaseCascadeModel(0.3, 0), [], y) == [1, 1, 1, 1, 1]
    assert click(BaseCascadeModel(0.3, 1), [], y) == [1, 1, 0, 1, 0]
    assert click(BaseCascadeModel(0.3, 3), [], y) == [0, 0, 0, 1, 1]
    assert click(BaseCascadeModel(0.3, 4), [], y) == [0, 0, 0, 1, 0]



import torch

from allrank.data.dataset_loading import PADDED_Y_VALUE
from allrank.models.losses import DEFAULT_EPS


def lambdaLoss(y_pred, y_true, eps=DEFAULT_EPS, padded_value_indicator=PADDED_Y_VALUE, weighing_scheme=None, k=None, sigma=1., mu=10.,
               reduction="sum", reduction_log="binary"):
    """
    LambdaLoss framework for LTR losses implementations, introduced in "The LambdaLoss Framework for Ranking Metric Optimization".
    Contains implementations of different weighing schemes corresponding to e.g. LambdaRank or RankNet.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param eps: epsilon value, used for numerical stability
    :param padded_value_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :param weighing_scheme: a string corresponding to a name of one of the weighing schemes
    :param k: rank at which the loss is truncated
    :param sigma: score difference weight used in the sigmoid function
    :param mu: optional weight used in NDCGLoss2++ weighing scheme
    :param reduction: losses reduction method, could be either a sum or a mean
    :param reduction_log: logarithm variant used prior to masking and loss reduction, either binary or natural
    :return: loss value, a torch.Tensor
    """
    device = y_pred.device
    y_pred = y_pred.clone()
    y_true = y_true.clone()

    padded_mask = y_true == padded_value_indicator
    y_pred[padded_mask] = float("-inf")
    y_true[padded_mask] = float("-inf")

    # Here we sort the true and predicted relevancy scores.
    y_pred_sorted, indices_pred = y_pred.sort(descending=True, dim=-1)
    y_true_sorted, _ = y_true.sort(descending=True, dim=-1)

    # After sorting, we can mask out the pairs of indices (i, j) containing index of a padded element.
    true_sorted_by_preds = torch.gather(y_true, dim=1, index=indices_pred)
    true_diffs = true_sorted_by_preds[:, :, None] - true_sorted_by_preds[:, None, :]
    padded_pairs_mask = torch.isfinite(true_diffs)

    if weighing_scheme != "ndcgLoss1_scheme":
        padded_pairs_mask = padded_pairs_mask & (true_diffs > 0)

    ndcg_at_k_mask = torch.zeros((y_pred.shape[1], y_pred.shape[1]), dtype=torch.bool, device=device)
    ndcg_at_k_mask[:k, :k] = 1

    # Here we clamp the -infs to get correct gains and ideal DCGs (maxDCGs)
    true_sorted_by_preds.clamp_(min=0.)
    y_true_sorted.clamp_(min=0.)

    # Here we find the gains, discounts and ideal DCGs per slate.
    pos_idxs = torch.arange(1, y_pred.shape[1] + 1).to(device)
    D = torch.log2(1. + pos_idxs.float())[None, :]
    maxDCGs = torch.sum(((torch.pow(2, y_true_sorted) - 1) / D)[:, :k], dim=-1).clamp(min=eps)
    G = (torch.pow(2, true_sorted_by_preds) - 1) / maxDCGs[:, None]

    # Here we apply appropriate weighing scheme - ndcgLoss1, ndcgLoss2, ndcgLoss2++ or no weights (=1.0)
    if weighing_scheme is None:
        weights = 1.
    else:
        weights = globals()[weighing_scheme](G, D, mu, true_sorted_by_preds)  # type: ignore

    # We are clamping the array entries to maintain correct backprop (log(0) and division by 0)
    scores_diffs = (y_pred_sorted[:, :, None] - y_pred_sorted[:, None, :]).clamp(min=-1e8, max=1e8)
    scores_diffs.masked_fill(torch.isnan(scores_diffs), 0.)
    weighted_probas = (torch.sigmoid(sigma * scores_diffs).clamp(min=eps) ** weights).clamp(min=eps)
    if reduction_log == "natural":
        losses = torch.log(weighted_probas)
    elif reduction_log == "binary":
        losses = torch.log2(weighted_probas)
    else:
        raise ValueError("Reduction logarithm base can be either natural or binary")

    if reduction == "sum":
        loss = -torch.sum(losses[padded_pairs_mask & ndcg_at_k_mask])
    elif reduction == "mean":
        loss = -torch.mean(losses[padded_pairs_mask & ndcg_at_k_mask])
    else:
        raise ValueError("Reduction method can be either sum or mean")

    return loss


def ndcgLoss1_scheme(G, D, *args):
    return (G / D)[:, :, None]


def ndcgLoss2_scheme(G, D, *args):
    pos_idxs = torch.arange(1, G.shape[1] + 1, device=G.device)
    delta_idxs = torch.abs(pos_idxs[:, None] - pos_idxs[None, :])
    deltas = torch.abs(torch.pow(torch.abs(D[0, delta_idxs - 1]), -1.) - torch.pow(torch.abs(D[0, delta_idxs]), -1.))
    deltas.diagonal().zero_()

    return deltas[None, :, :] * torch.abs(G[:, :, None] - G[:, None, :])


def lambdaRank_scheme(G, D, *args):
    return torch.abs(torch.pow(D[:, :, None], -1.) - torch.pow(D[:, None, :], -1.)) * torch.abs(G[:, :, None] - G[:, None, :])


def ndcgLoss2PP_scheme(G, D, *args):
    return args[0] * ndcgLoss2_scheme(G, D) + lambdaRank_scheme(G, D)


def rankNet_scheme(G, D, *args):
    return 1.


def rankNetWeightedByGTDiff_scheme(G, D, *args):
    return torch.abs(args[1][:, :, None] - args[1][:, None, :])


def rankNetWeightedByGTDiffPowed_scheme(G, D, *args):
    return torch.abs(torch.pow(args[1][:, :, None], 2) - torch.pow(args[1][:, None, :], 2))



import numpy as np
import torch

from allrank.data.dataset_loading import PADDED_Y_VALUE


def ndcg(y_pred, y_true, ats=None, gain_function=lambda x: torch.pow(2, x) - 1, padding_indicator=PADDED_Y_VALUE,
         filler_value=1.0):
    """
    Normalized Discounted Cumulative Gain at k.

    Compute NDCG at ranks given by ats or at the maximum rank if ats is None.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param ats: optional list of ranks for NDCG evaluation, if None, maximum rank is used
    :param gain_function: callable, gain function for the ground truth labels, e.g. torch.pow(2, x) - 1
    :param padding_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :param filler_value: a filler NDCG value to use when there are no relevant items in listing
    :return: NDCG values for each slate and rank passed, shape [batch_size, len(ats)]
    """
    idcg = dcg(y_true, y_true, ats, gain_function, padding_indicator)
    ndcg_ = dcg(y_pred, y_true, ats, gain_function, padding_indicator) / idcg
    idcg_mask = idcg == 0
    ndcg_[idcg_mask] = filler_value  # if idcg == 0 , set ndcg to filler_value

    assert (ndcg_ < 0.0).sum() >= 0, "every ndcg should be non-negative"

    return ndcg_


def __apply_mask_and_get_true_sorted_by_preds(y_pred, y_true, padding_indicator=PADDED_Y_VALUE):
    mask = y_true == padding_indicator

    y_pred[mask] = float('-inf')
    y_true[mask] = 0.0

    _, indices = y_pred.sort(descending=True, dim=-1)
    return torch.gather(y_true, dim=1, index=indices)


def dcg(y_pred, y_true, ats=None, gain_function=lambda x: torch.pow(2, x) - 1, padding_indicator=PADDED_Y_VALUE):
    """
    Discounted Cumulative Gain at k.

    Compute DCG at ranks given by ats or at the maximum rank if ats is None.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param ats: optional list of ranks for DCG evaluation, if None, maximum rank is used
    :param gain_function: callable, gain function for the ground truth labels, e.g. torch.pow(2, x) - 1
    :param padding_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: DCG values for each slate and evaluation position, shape [batch_size, len(ats)]
    """
    y_true = y_true.clone()
    y_pred = y_pred.clone()

    actual_length = y_true.shape[1]

    if ats is None:
        ats = [actual_length]
    ats = [min(at, actual_length) for at in ats]

    true_sorted_by_preds = __apply_mask_and_get_true_sorted_by_preds(y_pred, y_true, padding_indicator)

    discounts = (torch.tensor(1) / torch.log2(torch.arange(true_sorted_by_preds.shape[1], dtype=torch.float) + 2.0)).to(
        device=true_sorted_by_preds.device)

    gains = gain_function(true_sorted_by_preds)

    discounted_gains = (gains * discounts)[:, :np.max(ats)]

    cum_dcg = torch.cumsum(discounted_gains, dim=1)

    ats_tensor = torch.tensor(ats, dtype=torch.long) - torch.tensor(1)

    dcg = cum_dcg[:, ats_tensor]

    return dcg


def mrr(y_pred, y_true, ats=None, padding_indicator=PADDED_Y_VALUE):
    """
    Mean Reciprocal Rank at k.

    Compute MRR at ranks given by ats or at the maximum rank if ats is None.
    :param y_pred: predictions from the model, shape [batch_size, slate_length]
    :param y_true: ground truth labels, shape [batch_size, slate_length]
    :param ats: optional list of ranks for MRR evaluation, if None, maximum rank is used
    :param padding_indicator: an indicator of the y_true index containing a padded item, e.g. -1
    :return: MRR values for each slate and evaluation position, shape [batch_size, len(ats)]
    """
    y_true = y_true.clone()
    y_pred = y_pred.clone()

    if ats is None:
        ats = [y_true.shape[1]]

    true_sorted_by_preds = __apply_mask_and_get_true_sorted_by_preds(y_pred, y_true, padding_indicator)

    values, indices = torch.max(true_sorted_by_preds, dim=1)
    indices = indices.type_as(values).unsqueeze(dim=0).t().expand(len(y_true), len(ats))

    ats_rep = torch.tensor(data=ats, device=indices.device, dtype=torch.float32).expand(len(y_true), len(ats))

    within_at_mask = (indices < ats_rep).type(torch.float32)

    result = torch.tensor(1.0) / (indices + torch.tensor(1.0))

    zero_sum_mask = torch.sum(values) == 0.0
    result[zero_sum_mask] = 0.0

    result = result * within_at_mask

    return result
